import numpy as np
from shapely.ops import cascaded_union
from shapely.ops import unary_union
from shapely.geometry import Point
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from shapely.geometry import shape, JOIN_STYLE
import math


def is_neighbor(pos1,pos2,nx,ny):
 delta = 1e-3
 if abs(pos1[0] - pos2[0])<delta and abs(abs(pos1[1]-pos2[1]) -1) < delta:
  return True
 if abs(abs(pos1[0] - pos2[0]) -1)  < delta and abs(pos1[1] - pos2[1]) < delta:
  return True
 #if abs(abs(pos1[0] - pos2[0]) - (nx-1))  < delta and abs(pos1[1] - pos2[1]) < delta:
 # return True
 #if abs(pos1[0] - pos2[0])<delta and abs(abs(pos1[1]-pos2[1]) -(ny-1)) < delta:
 # return True

 return False


#Get an island of connecting pores----
def get_merged_pores(output,pos,total,nx,ny):
 for n,p in enumerate(pos):
  if not n in output and not n in total:
   for o in output:
    if is_neighbor(p,pos[o],nx,ny):
     output.append(n)
     get_merged_pores(output,pos,total,nx,ny)
     break
 #return output



def get_square(x,y,area):

 Na = 4
 r = math.sqrt(2.0*area/Na/math.sin(2.0 * math.pi/Na))
 dphi = 2.0*math.pi/Na;
 poly_clip = []
 for ka in range(Na+1):
      ph =  dphi/2 + (ka-1)*dphi
      px  = x + r * math.cos(ph)
      py  = y + r * math.sin(ph)
      poly_clip.append([px,py])
 return poly_clip

def merge_positions(pos,nx,ny):

 n_pores = len(pos)
 flat_merged_pores = []
 total_merged_pores = []
 n_filled = 0
 start = 0
 while n_filled < n_pores:
  pores = [start]
  get_merged_pores(pores,pos,flat_merged_pores,nx,ny)
  total_merged_pores.append(pores)
  flat_merged_pores = [element for tupl in total_merged_pores for element in tupl]
  n_filled = len(flat_merged_pores)
  #print(n_filled)
  #Get next available pores-------------
  while True:
    start +=1
    if not start in flat_merged_pores:
     break

 return total_merged_pores

def complement_positions(positions,nx,ny):

  compl = []
  for x in range(nx):
   for y in range(ny):
    included = False
    for tmp in positions:
     if tmp[0] == x and tmp[1] == y:
      included = True
      break
    if not included:
     compl.append([x,y])

  return compl



def  delete_unnecessary_points(polygons):

  #ELIMINATE DUPLICATE POINTS------
  n_del = 1
  n_iter = 0
  while n_del > 0:
   n_del = 0
   delete_list = []
   for n,poly in enumerate(polygons):
    tmp = []
    for p in range(len(poly)):
     p1 = np.array(poly[p])
     p2 = np.array(poly[(p+1)%(len(poly))])
     p3 = np.array(poly[(p+2)%(len(poly))])
     d1 = p2-p1
     d2 = p3-p2
     if not abs(np.dot(d1,d2))< 1e-2:
      tmp.append((p+1)%(len(poly)))
      n_del += 1
    delete_list.append(tmp)

   #print('Unnecessary_points: ' + str(n_del))
   #--------------------------------------
   new_polygons = []
   for d1 in range(len(delete_list)):
    tmp = []
    for p in range(len(polygons[d1])):
     if not p in delete_list[d1]:
      tmp.append(polygons[d1][p])
    new_polygons.append(tmp)
   polygons = new_polygons
   return polygons


def  delete_duplicate_points(polygons):

  #ELIMINATE DUPLICATE POINTS------
  n_del = 1
  while n_del > 0:
   n_del = 0
   delete_list = []
   for n,poly in enumerate(polygons):
    tmp = []
    for p in range(len(poly)):
     p1 = np.array(poly[p])
     p2 =np.array(poly[(p+1)%(len(poly))])
     if np.linalg.norm(p1-p2)<1e-3:
      tmp.append(p)
      n_del += 1
    delete_list.append(tmp)
   #print('Duplicate_points:' + str(n_del))
   #--------------------------------------
   for d1 in range(len(delete_list)):
    n = 0
    for d2 in delete_list[d1]:
     del polygons[d1][d2-n]
     n += 1
  return polygons


def GenerateRandomPoresGrid(argv):

  nx =  argv.setdefault('nx',10)
  ny =  argv.setdefault('ny',10)
  Np =  argv.setdefault('Np',40)
  positions = []
  for n_pore in range(Np):
    overlap = True
    while overlap:
     xr = np.random.randint(0,nx)
     yr = np.random.randint(0,ny)
     if not [xr,yr] in positions:
      positions.append([xr,yr])
      overlap = False


  return positions


def compute_inverse_polygons(positions,nx,ny,area,dx,dy,Lx,Ly,frame):

   compl = complement_positions(positions,nx+1,ny+1)
   merged_material = merge_positions(compl,nx,ny)
   coords = []
   for m1 in merged_material:
    p = []
    for m2 in m1:
     poly = get_square(compl[m2][0]*dx-Lx*0.5,compl[m2][1]*dy-Ly*0.5,area)
     p.append(Polygon(poly).buffer(0))
    m = MultiPolygon(p)
    #total = cascaded_union(m)

    total = cascaded_union(m).intersection(frame)
    #quit()
    external = list(total.exterior.coords)[0:-1]
    external = delete_duplicate_points(list([external]))
    external = delete_unnecessary_points(list(external))


    internal = []
    for k in total.interiors:
     tmp = list(k.coords)[0:-1]
     tmp = delete_duplicate_points([tmp])
     tmp = delete_unnecessary_points(tmp)
     internal.append(tmp[0])
    coords.append({'external':external[0],'internal':internal})

   return coords



def compute_polygons_from_grid(values):
  #Convert--
  (nx,ny)=np.shape(values)
  (ix,iy) = values.nonzero()
  positions_old = []
  for a,b in zip(ix,iy):
   positions_old.append([a,b])


  #positions_old = argv['x']
  #positions_old = np.reshape(argv['x'],(len(argv['x'])/2,2))

  #Lx =  argv.setdefault('Lx',40.0)
  Lx = 1.0
  Ly = 1.0
  #Ly =  argv.setdefault('Ly',40.0)
  #phi =  argv.setdefault('porosity',0.3)
  Na = 4
  #nx =  argv.setdefault('nx',10)
  #ny =  argv.setdefault('ny',10)
  dx = Lx/nx
  dy = Ly/ny
  area = dx*dy
  frame_tmp = []
  frame_tmp.append([float(-Lx)/2,float(Ly)/2])
  frame_tmp.append([float(Lx)/2,float(Ly)/2])
  frame_tmp.append([float(Lx)/2,float(-Ly)/2])
  frame_tmp.append([float(-Lx)/2,float(-Ly)/2])
  frame = Polygon(frame_tmp)



  #quit()
  delta = 1e-3

  positions = []
  polygons = []
  for pos in positions_old:
    xr = pos[0]
    yr = pos[1]
    polygons.append(get_square(xr*dx-Lx*0.5,\
                               yr*dy-Lx*0.5,\
                               area))
    positions.append(pos)

    #if argv.setdefault('add_periodic',True):
    if 1 ==1:
    #ADD PERIODIC PORES--------------------------
     if xr == 0:
       polygons.append(get_square(xr*dx-Lx*0.5 + Lx, yr*dy-Lx*0.5,area))
       positions.append([nx,yr])

     if yr == 0:
      polygons.append(get_square(xr*dx-Lx*0.5     , yr*dy-Lx*0.5 + Ly ,area))
      positions.append([xr,ny])

     if yr == 0 and xr == 0:
      polygons.append(get_square(xr*dx-Lx*0.5 + Lx, yr*dy-Lx*0.5 + Ly ,area))
      positions.append([nx,ny])

    #-----------------------------------------------------

  #print('START')
  lone_pores = []
  #if argv.setdefault('eliminate_lone_pores',True):
  if 1 == 1:
  #if argv.setdefault('eliminate_lone_pores',True):
   #Eliminate holes-----
   th = 15
   compl = complement_positions(np.array(positions),nx,ny)
   merged_material = merge_positions(compl,nx,ny)
   #print('Number of holes:')
   for hole in merged_material:
    if len(hole) < th:
     for pos in hole:
      #fill missing pores
      positions.append(compl[pos])
      polygons.append(get_square(compl[pos][0]*dx-Lx*0.5,\
                                compl[pos][1]*dy-Ly*0.5,\
                                area))
      lone_pores.append(len(polygons)-1)


      if 1 == 1:
      #if argv.setdefault('add_periodic',True):

       if compl[pos][0] == 0:
         polygons.append(get_square(compl[pos][0]*dx-Lx*0.5+Lx,\
                                    compl[pos][1]*dy-Ly*0.5,\
                                    area))
         lone_pores.append(len(polygons)-1)

         positions.append([nx,compl[pos][1]])

       if compl[pos][1] == 0:
        polygons.append(get_square(compl[pos][0]*dx-Lx*0.5,\
                                   compl[pos][1]*dy-Ly*0.5+Ly,\
                                   area))
        lone_pores.append(len(polygons)-1)
        positions.append([compl[pos][0],ny])

       if compl[pos][0] == 0 and compl[pos][1] == 0:
        lone_pores.append(len(polygons)-1)
        polygons.append(get_square(compl[pos][0]*dx-Lx*0.5 + Lx,\
                                   compl[pos][1]*dy-Ly*0.5 + Ly,\
                                   area))
        lone_pores.append(len(polygons)-1)
        positions.append([nx,ny])
  #print('Added lone pores:' + str(len(lone_pores)))
  #quit()
  #MERGE POLYGONS:#---------
  merge_poly = True
  if merge_poly:
   polygons = np.array(polygons)
   merged = merge_positions(positions,nx,ny)

   npp = 0
   polygons_final = []
   for m1 in merged:
    p = []
    for m2 in m1:
     #p.append(Polygon(polygons[m2]).buffer(1e-6))
     tt = Polygon(polygons[m2]).buffer(1e-6, resolution=1,mitre_limit=20.0,join_style=2)
     p.append(tt)
    m = MultiPolygon(p)
    npp += len(p)
    total = cascaded_union(m)


    polygons_final.append(list(total.exterior.coords)[0:-1])

  else:
   polygons_final = polygons
   merged = []
   for p in range(len(polygons_final)):
    merged.append([p])

  #print(npp)
  #print(len(positions))
  #polygons_final = polygons
  #type it up--------
  polygons_final = delete_duplicate_points(list(polygons_final))
  polygons_final = delete_unnecessary_points(list(polygons_final))
  #------------------

  #inverse_polygons = compute_inverse_polygons(positions,nx,ny,area,dx,dy,Lx,Ly,frame)
  #inverse_polygons = delete_duplicate_points(list(inverse_polygons))
  #inverse_polygons = delete_unnecessary_points(list(inverse_polygons))
  #--------------------------------------------------
  #----------------------------
  area = 0
  for p in polygons_final:
   area += Lx*Ly - frame.difference(Polygon(p)).area
  porosity = area/Lx/Ly

  delta = 1e-3
  #output = {'polygons':polygons_final,'Lx':Lx*(1.0-delta),'Ly':Ly*(1.0-delta),'test':merged,'lone':lone_pores,'porosity':porosity,'positions':positions,'inverse_polygons':inverse_polygons}

  polygons = []
  for poly in polygons_final:
   tmp = []
   for p in poly:
    tmp.append(np.array(p))
   polygons.append(np.array(tmp))


  return np.array(polygons)



def GenerateRandomPoresGrid(**options):

 nx = options['nx']
 ny = options['ny']
 Np = options['np']

 values = np.zeros((nx,ny))

 pos = []
 while len(pos) < Np :
  x = np.random.randint(0,nx)
  y = np.random.randint(0,ny)
  tmp = [x,y]
  if not tmp in pos:
   pos.append(tmp)
   values[x,y] = 1

 #values = list(np.reshape(pos,(1,2*Np))[0])

 polygons = compute_polygons_from_grid(values)
 return values,polygons



def generate_first_guess(options):

 Lx = options['Lx']
 Ly = options['Ly']
 Np = options['Np']

 pos = []
 for n in range(Np):
  x = np.random.uniform(-Lx/2.0,Lx/2.0)
  y = np.random.uniform(-Ly/2.0,Ly/2.0)
  pos.append(x)
  pos.append(y)

 return pos





def compute_cost(options,pos):

 output = adapt_data(options,pos)

 return output['polygons']



def plot_sample(options,pos):

 output = adapt_data(options,pos)


 #positions = prepare_data(pos,options)
 #options.update({'eliminate_lone_pores':True})
 #options.update({'add_periodic':True})
 Lx = options['Lx']
 Ly = options['Ly']
 #output = adapt_data(options,positions)
 coords = output['polygons']


 #test = output['test']
 #lone = output['lone']



 #r = lambda: float(random.randint(0,255))/255.0
 #colors = []
 #for g in test:
 # colors.append((r(),r(),r()))



 fig = figure()
 ax = fig.add_subplot(111, aspect='equal')
 frame = []
 frame.append([float(-Lx)/2,float(Ly)/2])
 frame.append([float(Lx)/2,float(Ly)/2])
 frame.append([float(Lx)/2,float(-Ly)/2])
 frame.append([float(-Lx)/2,float(-Ly)/2])
 frame.append([float(-Lx)/2,float(Ly)/2])

 for f,p in enumerate(frame):
   plot([p[0],frame[(f+1)%len(frame)][0]],[p[1],frame[(f+1)%len(frame)][1]],color='g',linewidth=4)

 path = create_path(frame)
 patch = patches.PathPatch(path, facecolor='orange', lw=2,alpha=0.5)
 ax.add_patch(patch)
 #--------------------------------------------------------------
 #plot polygons---i
 for n,poly in enumerate(coords):
  color = 'white'
  #for t,tt in enumerate(test):
  # if n in tt:
  #  color = colors[t]
  #  break
  #if n in lone:
  # color = 'black'
  #for p in range(len(poly)):
  # p1 = np.array(poly[p])
  # p2 = np.array(poly[(p+1)%(len(poly)-1)])
  # if np.linalg.norm(p1-p2)<1e-4:
  #  print(p1)
  #  print(p2)
  #  print(" ")

  p = np.array(poly).copy()
  path = create_path(p)
  patch = patches.PathPatch(path, facecolor=color, lw=2)
  ax.add_patch(patch)



 axis('off')
 axis('equal')
 ylim([-Lx/2,Lx/2])
 xlim([-Ly/2,Ly/2])

 show()
 return fig


def plot_coordinates(output):

 coords = output['coords']
 Lx = output['Lx']
 Ly = output['Ly']

 #P = [[0,0],[0,Ly],[Lx,Ly],[Lx,0],[0,-Ly],[-Lx,0],[-Lx,-Ly],[-Lx,Ly],[Lx,-Ly]]
 P = [[0,0]]
 for coord in coords:
  Np=len(coord)
  for tr in P:
   for ka in range(len(coord)-1):
    plot([coord[ka][0]+tr[0],coord[(ka+1)][0] + tr[0]],\
         [coord[ka][1]+tr[1],coord[(ka+1)][1] + tr[1]],color='b')
    #if Np > 4:
    #print(Np)
    #text(coord[ka][0]+tr[0],coord[ka][1]+tr[1],str(ka))

 plot([-Lx/2,-Ly/2],[-Lx/2,Ly/2],color='r',ls='--',lw=3)
 plot([-Lx/2,Ly/2],[Lx/2,Ly/2],color='r',ls='--',lw=3)
 plot([Lx/2,Ly/2],[Lx/2,-Ly/2],color='r',ls='--',lw=3)
 plot([Lx/2,-Ly/2],[-Lx/2,-Ly/2],color='r',ls='--',lw=3)

 axis('equal')
 axis('off')
 xlim([-Lx,Lx])
 ylim([-Ly,Ly])
 show()


def plot_configuration(output):

 vector = output['vector']


 L = 40
 P = [[0,0],[0,L],[L,L],[L,0],[0,-L],[-L,0],[-L,-L],[-L,L],[L,-L]]
 r = 5.0*sqrt(0.4)
 pairs = []
 N = len(vector)/2
 for n in range(N):
  x = vector[2*n]
  y = vector[2*n+1]
  for tr in P:
   pairs = []
   for ka in range(4):
    ph =  (ka-1) * np.pi/2.0
    px  = x + tr[0] + r * math.cos(ph+np.pi/4.0)
    py  = y  + tr[1] + r * math.sin(ph+np.pi/4.0)
    pairs.append([px,py])
   for ka in range(4):
    plot([pairs[ka][0],pairs[(ka+1)%4][0]],[pairs[ka][1],pairs[(ka+1)%4][1]],color='b')
 plot([-L/2,-L/2],[-L/2,L/2],color='r',ls='--',lw=3)
 plot([-L/2,L/2],[L/2,L/2],color='r',ls='--',lw=3)
 plot([L/2,L/2],[L/2,-L/2],color='r',ls='--',lw=3)
 plot([L/2,-L/2],[-L/2,-L/2],color='r',ls='--',lw=3)

 axis('equal')
 axis('off')
 xlim([-L,L])
 ylim([-L,L])
 show()



if __name__ == "__main__":


 vector = compute_first_guess()

 kappa = compute_kappa(vector)








import xml.etree.ElementTree as ET
import os,sys
import numpy as np
import random
import math
from matplotlib.pylab import *
import subprocess

def create_structured_bulk(argv):



 res = argv.setdefault('reservoirs',False)
 direction = argv['direction']


 Lx = argv['lx']
 Ly = argv['ly']
 mesh_ext = argv['step']
 layers_x = int(Lx/mesh_ext)
 layers_y = int(Ly/mesh_ext)
 store = open("mesh.geo", 'w+')
 store.write( 'Point(0) = {' + str(-Lx/2.0) +','+ str(-Ly/2.0)+',0,'+ str(mesh_ext) +'};\n')
 store.write('out[] = Extrude{0,''' + str(Ly) + r''',0}{ Point{0};Layers{''' + str(layers_x) + r'''};Recombine;};''' + '\n')
 store.write('out[] = Extrude{' + str(Lx) + r''',0,0}{ Line{out[0]};Layers{''' + str(layers_y) + r'''};Recombine;};''' + '\n')

 #Create Surface
 strc = r'''Physical Surface('bulk') = {out[1]};'''+'\n'
 store.write(strc)


 bs = []
 if argv.setdefault('Periodic',[True,True,True])[1] :
  if res and direction=='x':
   strc = r'''Physical Line('Hot') = {1};''' + '\n'
   store.write(strc)
   strc = r'''Physical Line('Cold') = {2};''' +'\n'
   store.write(strc)
  else:
   strc = r'''Physical Line('Periodic_1') = {1};''' + '\n'
   store.write(strc)
   strc = r'''Physical Line('Periodic_2') = {2};''' + '\n'
   store.write(strc)
 else:
  bs.append(1)
  bs.append(3)

 if argv.setdefault('Periodic',[True,True,True])[0] :
  if res and direction=='y':
   strc = r'''Physical Line('Cold') = {3};''' + '\n'
   store.write(strc)
   strc = r'''Physical Line('Hot') = {4};''' +'\n'
   store.write(strc)
  else:
   strc = r'''Physical Line('Periodic_3') = {3};''' + '\n'
   store.write(strc)
   strc = r'''Physical Line('Periodic_4') = {4};''' +'\n'
   store.write(strc)

 else:
  bs.append(2)
  bs.append(4)

 if len(bs) > 0:
  strc = r'''Physical Line('Boundary') = {'''
  for p,side in enumerate(bs) :
   strc +=str(side)
   if p == len(bs)-1:
    strc += '};\n'
   else :
    strc += ','
  store.write(strc)

 strc = 'Periodic Line{1}={-3};\n'
 store.write(strc)
 strc = 'Periodic Line{2}={-4};\n'
 store.write(strc)

 store.close()

def create_unstructured_bulk(argv):

  res = argv.setdefault('reservoirs',False)
  direction = argv['direction']

  Lx = float(argv['lx'])
  Ly = float(argv['ly'])
  frame = []

  frame.append([-Lx/2,Ly/2])
  frame.append([Lx/2,Ly/2])
  frame.append([Lx/2,-Ly/2])
  frame.append([-Lx/2,-Ly/2])


  mesh_ext = argv['step']
  points = []
  lines = []

  store = open("mesh.geo", 'w+')

  for k,p in enumerate(frame) :
   store.write( 'Point('+str(k) +') = {' + str(p[0]) +','+ str(p[1])+',0,'+ str(mesh_ext) +'};\n')

  ll = 0
  for k,p in enumerate(frame) :
   p1 = k
   p2 = (k+1)%len(frame)
   ll += 1
   store.write( 'Line('+str(ll) +') = {' + str(p1) +','+ str(p2)+'};\n')

  strc = 'Line Loop(0) = {'
  for p,point in enumerate(frame) :
   strc +=str(p+1)
   if p == len(frame)-1:
    strc += '};\n'
   else :
    strc += ','
  store.write(strc)

  #Create Surface
  strc = 'Plane Surface(0) = {0};\n'
  store.write(strc)
  strc = r'''Physical Surface('Matrix') = {0};'''+'\n'
  store.write(strc)

  bs = []
  if argv.setdefault('Periodic',[True,True,True])[1] :

   if res and direction=='y':
    strc = r'''Physical Line('Cold') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Hot') = {4};''' +'\n'
   else:
    strc = r'''Physical Line('Periodic_1') = {1};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Periodic_2') = {3};''' + '\n'
    store.write(strc)
  else:
   bs.append(1)
   bs.append(3)

  if argv.setdefault('Periodic',[True,True,True])[0] :

   if res and direction=='x':
    strc = r'''Physical Line('Cold') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Hot') = {4};''' +'\n'
    store.write(strc)
   else:
    strc = r'''Physical Line('Periodic_3') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Periodic_4') = {4};''' +'\n'
    store.write(strc)
  else:
   bs.append(2)
   bs.append(4)


  if len(bs) > 0:
   strc = r'''Physical Line('Boundary') = {'''
   for p,side in enumerate(bs) :
    strc +=str(side)
    if p == len(bs)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)


  strc = 'Periodic Line{1}={-3};\n'
  store.write(strc)
  strc = 'Periodic Line{2}={-4};\n'
  store.write(strc)

#-------------------------------------------------------
  store.close()




def mesh(argv):

  if argv.setdefault('unstructured',True):
   create_unstructured_bulk(argv)
  else:
   create_structured_bulk(argv)



from __future__ import print_function
get_ipython().magic(u'matplotlib nbagg')
from openbte.material import *
from openbte.plot import *
from ipywidgets import interact, interactive, fixed, interact_manual,HBox,VBox
import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Button, GridBox, Layout, ButtonStyle
from IPython.display import Latex
import numpy as np
#from openbte.solver_new import *
#from openbte.geometry import *


t = widgets.Text(value = 'Thermal Conductivity [W/m/K]: ')

sel = widgets.Select(
    options=['Geometry',' '],
    value='Geometry',
    description='Variable',
    disabled=False
)

w = widgets.FloatSlider(
    value=0.2,
    min=0.01,
    max=0.3,
    step=0.01,
    description='Porosity:',
    disabled=False,
    continuous_update=False,
    orientation='horizontal',
    readout=True,
    readout_format='.3f')
w.style.handle_color = 'lightblue'


ki = widgets.BoundedFloatText(
    value=100.0,
    min=0,
    max=1000,
    step=0.1,
    description='TC Inclusion [W/K/m]',
    disabled=True,
    layout = {'width': '300px'},\
    style = {'description_width': '150px'})

mi = widgets.BoundedFloatText(
    value=0.1,
    min=1e-3,
    max=1e3,
    step=0.001,
    description='Kn Inclusion',
    disabled=True,
    layout = {'width': '300px'},\
    style = {'description_width': '150px'})




def create_geometry(porosity,shape,angle,lattice,dx,dy,inclusion,direction,xp,yp):

    if lattice == 'squared':
     type = 'porous/square_lattice'
     step = 0.1
    if lattice == 'staggered':
     type = 'porous/staggered_lattice'
     step = 0.2
    if lattice == 'honeycomb':
      type = 'porous/hexagonal_lattice'
      step = 0.1 *np.sqrt(sqrt(3.0))

    geo = Geometry(type = type,shape=shape,
                                  lx = 1.0, ly = 1.0,\
                                  step=step/1.0,\
                                  angle=angle,\
                                  dx = dx,\
                                  dy = dy,\
                                  porosity=porosity,
                                  direction=direction,
                                  inclusion = inclusion,
                                  Periodic=[xp,yp,False],
                                  mesh=True,plot=False);


    sel.options=['Geometry',' ']
    sel.value = ' '
    sel.value = 'Geometry'
    t.value = 'Thermal Conductivity [W/m/K]: '

    
    mi.disabled = not inclusion
    ki.disabled = not inclusion
    
        


def create_material(kappa_matrix,kappa_inclusion,mfp_matrix,mfp_inclusion):
    Material(region='Matrix',kappa = kappa_matrix,filename='material_a',mfps=[mfp_matrix*1e-9]);
    Material(region='Inclusion',kappa = kappa_inclusion,filename='material_b',mfps=[mfp_inclusion*1e-9]);
    w.value = w.value


def plot_data(variable):

 if not variable in ['Geometry',' '] :
  if variable == 'Flux (Magnitude) [Fourier]':
      var = 'fourier_flux'
      direction = 'magnitude'
  elif variable == 'Flux (X) [Fourier]':
      var = 'fourier_flux'
      direction = 'x'
  elif variable == 'Flux (Y) [Fourier]':
      var = 'fourier_flux'
      direction = 'y'
  elif variable == 'Temperature [Fourier]':
      var = 'fourier_temperature'
      direction = None

  if variable == 'Flux (Magnitude) [BTE]':
    var = 'bte_flux'
    direction = 'magnitude'
  elif variable == 'Flux [BTE]':
    var = 'bte_flux'
    direction = 'x'
  elif variable == 'Flux (Y) [BTE]':
    var = 'bte_flux'
    direction = 'y'
  elif variable == 'Temperature [BTE]':
    var = 'bte_temperature'
    direction = None



  Plot(variable='map/' + var,direction=direction,show=True,write=False,streamlines=False,repeat_x=1,repeat_y =1,plot_interfaces=True)
  #Plot(variable='line/'+ var,direction=direction,show=True,write=False,streamlines=False,repeat_x=1,repeat_y =1,plot_interfaces=True)


 if variable == 'Geometry':
  Geometry(type='load',filename = 'geometry',plot=True)

def run_fourier(b):
 sol = Solver(max_bte_iter = 0,verbose=False);

 kappa = sol.state['kappa_fourier']
 #if b.value==0:
 # display(t)
 t.value = 'Thermal Conductivity [W/m/K]: '.ljust(20) +  '{:8.2f}'.format(kappa)
 b.value +=1
 sel.options=['Geometry',display(Latex(r'''Flux (X) [Fourier]''')), 'Flux (Y) [Fourier]', 'Flux (Magnitude) [Fourier]','Temperature [Fourier]']

 sel.value = 'Flux (Magnitude) [Fourier]'

def run_bte(b):
  sol = Solver(max_bte_iter = 100,verbose=True);

  kappa = sol.state['kappa_bte']
  #if b.value==0:
  # display(t)
  t.value = 'Thermal Conductivity [W/m/K]: '.ljust(20) +  '{:8.2f}'.format(kappa)
  b.value +=1
  sel.options=['Geometry','Flux [X] [BTE]', 'Flux (Y) [BTE]', 'Flux (Magnitude) [BTE]','Temperature [BTE]']

  sel.value = 'Flux (Magnitude) [BTE]'



w = widgets.FloatSlider(
    value=0.2,
    min=0.01,
    max=0.3,
    step=0.01,
    description='Porosity:',
    disabled=False,
    continuous_update=False,
    orientation='horizontal',
    readout=True,
    readout_format='.3f')
w.style.handle_color = 'lightblue'


d = widgets.RadioButtons(
    options=['circle', 'square', 'triangle'],
    value='square',
    description='shape:',
    disabled=False
)

l = widgets.RadioButtons(
    options=['squared', 'staggered', 'honeycomb'],
    value='squared',
    description='Pore lattice:',
    disabled=False
)

a = widgets.FloatSlider(
    value=0.0,
    min=0.0,
    max=360,
    step=10,
    description='Angle:',
    disabled=False,
    continuous_update=False,
    orientation='horizontal',
    readout=True,
    readout_format='.0f')
a.style.handle_color = 'lightblue'

dx = widgets.FloatSlider(
    value=0.0,
    min=-0.5,
    max=0.5,
    step=0.1,
    description='dx',
    disabled=False,
    continuous_update=False,
    orientation='horizontal',
    readout=True,
    readout_format='.1f')
dx.style.handle_color = 'red'

dy = widgets.FloatSlider(
    value=0.0,
    min=-0.5,
    max=0.5,
    step=0.1,
    description='dy',
    disabled=False,
    continuous_update=False,
    orientation='horizontal',
    readout=True,
    readout_format='.1f')
dy.style.handle_color = 'red'


inc = widgets.Checkbox(
    value=True,
    description='Inclusion',
    disabled=False
)


xp = widgets.Checkbox(
    value=True,
    description='Periodic along x',
    disabled=False
)

yp = widgets.Checkbox(
    value=True,
    description='Periodic along y',
    disabled=False
)


gra = widgets.RadioButtons(
    options=['x', 'y'],
    value='x',
    description='Direction of the applied temperature',
    disabled=False
)

km = widgets.BoundedFloatText(
    value=150.0,
    min=1,
    max=1000,
    step=0.1,
    description='TC Matrix [W/K/m]',
    disabled=False,
    layout = {'width': '300px'},\
    style = {'description_width': '150px'})



mm = widgets.BoundedFloatText(
    value=0.1,
    min=1e-3,
    max=1e3,
    step=0.001,
    description='Kn Matrix',
    disabled=False,
    layout = {'width': '300px'},\
    style = {'description_width': '150px'})



#plot geometry-------
test = interact(create_geometry, porosity=w, shape=d, angle=a,lattice=l,dx=dx,dy=dy,inclusion=inc,direction=gra,xp=xp,yp=yp);
v3 = interactive(create_material, kappa_matrix=km,kappa_inclusion=ki,mfp_matrix = mm,mfp_inclusion=mi);



#display(test.children[:])
#v1 = VBox(test.children[0:2])
#v2 = VBox(test.children[2:4])
#display(HBox([v1,v2,v3]))
display(v3)
#--------------------


v = interactive(plot_data,variable=sel);
display(v)


#plot output-------
button_fourier = widgets.Button(description="Fourier");
button_fourier.value = 0
button_fourier.on_click(run_fourier)
button_bte= widgets.Button(description="BTE");
button_bte.on_click(run_bte)
button_bte.value = 0
display(HBox([button_fourier,button_bte,t]))
#-------------------





#



import numpy as np
import math
from shapely.geometry import Point
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

def GenerateCustomPores(argv):
    
  Lx = float(argv['lx'])
  Ly = float(argv['ly'])
  frame_tmp = []
  frame_tmp.append([float(-Lx)/2,float(Ly)/2])
  frame_tmp.append([float(Lx)/2,float(Ly)/2])
  frame_tmp.append([float(Lx)/2,float(-Ly)/2])
  frame_tmp.append([float(-Lx)/2,float(-Ly)/2])
  frame = Polygon(frame_tmp)

  #Periodicity------------
  pbc = []
  pbc.append([0,0])
  if argv.setdefault('automatic_periodic',True):
   pbc.append([Lx,0])
   pbc.append([-Lx,0])
   pbc.append([0,Ly])
   pbc.append([0,-Ly])
   pbc.append([Lx,Ly])
   pbc.append([-Lx,-Ly])
   pbc.append([-Lx,Ly])
   pbc.append([Lx,-Ly])
   #---------------------

  if 'polygons' in argv:
   polygons = np.array(argv['polygons'])
   for poly in polygons:
    for p in poly:
     p[0] *=Lx   
     p[1] *=Ly
     #p[0] -=Lx/2.0
     #p[1] -=Ly/2.0
  else:
   tmp = np.loadtxt(argv['polyfile'])
   if len(np.shape(tmp)) == 1:
    N = 1
    M = len(tmp)
   else:
    (N,M) = np.shape(tmp)
   polygons = tmp.reshape((N,int(M/2),2))


  final = []
  for poly in polygons:
    for kp in range(len(pbc)):
     tmp = []
     for p in poly:
      cx = p[0] + pbc[kp][0]
      cy = p[1] + pbc[kp][1]
      tmp.append([cx,cy])
     p1 = Polygon(tmp)
     if p1.intersects(frame): 
      final.append(tmp)

  return np.array(frame_tmp),np.array(final)
  





from matplotlib.pylab import *

import os,sys


LARGE = 30
c1 = '#1f77b4'
c2 = '#f77f0e'
c3 = '#2ca02c'


def savefigure(prefix = './'):


 namefile =  prefix + sys.argv[0].split('.')[0]+'.png'

 savefig(namefile,dpi=300)




def init_plotting(extra_x_padding = 0.0,extra_y_padding=0.0,extra_bottom_padding = 0.0,extra_top_padding=  0.0,paraview=False,square=True,delta_square = 0,presentation=False):

 rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
 rc('font',**{'family':'serif','sans-serif':['Computer Modern Sans serif']})

 rcParams['xtick.major.pad']='10'
 rc('text', usetex=True)

 rcParams['lines.linewidth'] = 4
 rcParams['font.size'] = LARGE
 rcParams['xtick.labelsize'] = LARGE
 rcParams['ytick.labelsize'] = LARGE
 if square:
  mpl.rcParams['figure.figsize'] = [4.0, 4.0]
 else:
  mpl.rcParams['figure.figsize'] = [8.0, 5.0]
 if presentation:
  mpl.rcParams['figure.figsize'] = [16.0, 12.0]

 rcParams['text.latex.preamble'] = [r'\usepackage{hyperref}',]


 mpl.rcParams['figure.dpi'] = 80
 mpl.rcParams['savefig.dpi'] = 300
 rcParams['xtick.major.size'] = 3
 rcParams['xtick.minor.size'] = 3
 rcParams['figure.edgecolor'] = 'k'
 rcParams['figure.facecolor'] = 'w'
 rcParams['xtick.major.width'] = 1
 rcParams['xtick.minor.width'] = 1
 rcParams['ytick.major.size'] = 3
 rcParams['ytick.minor.size'] = 3
 rcParams['ytick.major.width'] = 1
 rcParams['ytick.minor.width'] = 1
 rcParams['legend.frameon'] = False
 rcParams['legend.fontsize'] = 25
 rcParams['axes.linewidth'] = 1
 #axes([0.15+extra_x_padding,0.20+extra_bottom_padding,0.78-extra_x_padding-extra_y_padding,0.75-extra_bottom_padding-extra_top_padding])
 axes([0,0,1.0,1.0])



from __future__ import absolute_import
import numpy as np

from .compute_dom import compute_dom_2d
from mpi4py import MPI
import os
import os.path
import deepdish as dd
#from matplotlib.pylab import *

class Material(object):

 def __init__(self,**argv):

  #defaults--
  #argv.setdefault('grid',[50,12,24])

  #MFP discretization-----------------------------
  self.state = {}
 
  argv.setdefault('model','nongray')
  if argv['model'] == 'load':
   #MPI.COMM_WORLD.Barrier()
   if MPI.COMM_WORLD.Get_rank() == 0:
    data = dd.io.load(argv.setdefault('filename','material.hdf5'))
   else: data=None
   self.state.update(MPI.COMM_WORLD.bcast(data,root=0))
   #print(self.state['region'])
  else:
   #if MPI.COMM_WORLD.Get_rank() == 0:
   dom = compute_dom_2d(argv)

   if argv['model'] == 'nongray':
     argv.update({'dom':dom})   
     self.import_data_new(argv)

   
   if argv['model'] == 'gray':
    self.create_gray_model(argv)
   #-----------------------------------------------

   #Angular discretization-------------------------
   if MPI.COMM_WORLD.Get_rank() == 0:
    region = argv.setdefault('region','Matrix')
    data = {'dom':dom,'region':region}

   else: data=None
   self.state.update(MPI.COMM_WORLD.bcast(data,root=0))

   #-----------------------------------------------
     #SAVE FILE--------------------
   if argv.setdefault('save',True):
    if MPI.COMM_WORLD.Get_rank() == 0:
     dd.io.save(argv.setdefault('filename','material') + '.hdf5', self.state)


 #def get_region(self):
     #return self.region

 def create_gray_model(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:
    data = {}
    mfp = argv.setdefault('mfps',[100e-9])
    n_mfp = len(mfp)
    kappa = argv.setdefault('kappa',1.0)

    data.update({'kappa_bulk_tot':kappa})
    data.update({'mfp_bulk':mfp})
    data.update({'kappa_bulk':[kappa]})
    data.update({'B0':np.array(n_mfp*[1.0/n_mfp])})

    if 'special_point' in argv.keys():
     sp = argv['special_point']
     tmp = np.zeros(n_mfp)
     tmp[sp] = 1.0
     tmp = np.tile(tmp,[n_mfp,1])
     data.update({'B1':tmp})
     data.update({'B2':tmp})
    else:
     data.update({'B1':np.eye(n_mfp)})
     data.update({'B2':np.eye(n_mfp)})

    data.update({'mfp_sampled':mfp})
  else: data=None
  self.state = MPI.COMM_WORLD.bcast(data,root=0)



 def import_data_new(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:

    
    dom = argv['dom']  
    if '/' in argv['matfile']:
     filename = argv['matfile']
    else:
     filename = os.path.dirname(__file__) + '/materials/'+ argv['matfile']
  
    tmp = np.loadtxt(filename)
    mfp_bulk = tmp[:,0]*1e9
    kappa_bulk = tmp[:,1]
    ftheta = dom['ftheta']
    dtheta = dom['dtheta']
    min_rmfp = 0.99*min(mfp_bulk)*min(ftheta)
    max_rmfp = 1.01*max(mfp_bulk)*max(ftheta)
    rmfp_log = np.linspace(np.log10(min_rmfp*0.99),np.log10(max_rmfp*1.01),argv['n_rmfp'])
    #---DISCRETIZE MFPS in LOG SPACE---------------------
    rmfp = []
    for m in rmfp_log:
     rmfp.append(pow(10,m))
    rmfp = np.array(rmfp)
    
    trials = np.outer(mfp_bulk,ftheta)    
    coeff1 = kappa_bulk/mfp_bulk
    J0=np.outer(np.ones(len(mfp_bulk)),dtheta)
    J1 = np.outer(coeff1/sum(coeff1),ftheta*dtheta)   
    coeff2 = kappa_bulk/mfp_bulk/mfp_bulk
    J2 = np.outer(coeff2/sum(coeff2),dtheta) #Here we should multiply by 2 for the symmetry and divide by 2 for the average
    J3 = np.outer(1/mfp_bulk,ftheta*dtheta)   
    
    data = {}
    data.update({'kappa_bulk_tot':sum(kappa_bulk)})
    data.update({'kappa_bulk':kappa_bulk})
    data.update({'mfp_bulk':mfp_bulk})
    data.update({'trials':trials})
    data.update({'rmfp':rmfp})
    data.update({'J0':J0})
    data.update({'J1':J1})
    data.update({'J2':J2})
    data.update({'J3':J3})  
    
  else: data=None
  self.state = MPI.COMM_WORLD.bcast(data,root=0)
  



 def import_data(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:

    if '/' in argv['matfile']:
     filename = argv['matfile']
    else:
     filename = os.path.dirname(__file__) + '/materials/'+ argv['matfile']
    #-------
    data = {}

    n_rmfp = int(argv['n_rmfp'])
    
    #READ DATA-------------------------------------------
    tmp = np.loadtxt(filename)
    mfp_bulk = tmp[:,0]
    kappa_bulk = tmp[:,1]
    n_mfp_bulk = len(mfp_bulk)
    
    #experimental----------
    ftheta = argv['dom']['ftheta']
    #dtheta_ave = argv['dom']['dtheta_ave']
    #dtheta = argv['dom']['dtheta']
    #n_theta = len(dtheta_ave)
    #---------------------------------
    
    min_rmfp = 0.99*min(mfp_bulk)*min(ftheta)
    max_rmfp = 1.01*max(mfp_bulk)*max(ftheta)
    rmfp_log = np.linspace(np.log10(min_rmfp*0.99),np.log10(max_rmfp*1.01),n_rmfp)
    #---DISCRETIZE MFPS in LOG SPACE---------------------
    rmfp = []
    for m in rmfp_log:
     rmfp.append(pow(10,m))
    
    #--------------------------------------------------------
    #MFP interpolation---------------------------
   # B0 = np.zeros(n_rmfp)
    #B1 = np.zeros(n_rmfp)
    #J = np.zeros(n_mfp_bulk)
    
    #B2 = np.zeros(n_rmfp)
    #factor = 0
    #kappa_tot = 0
    #for m in range(n_mfp_bulk):
    # kappa = kappa_bulk[m]
    # mfp = mfp_bulk[m]
    # a = kappa/mfp
    # J[m] = a
    # factor += a
    # for t in range(n_theta) :   
    #  [n1,n2,f1,f2] = self.get_mfp_interpolation(rmfp_log,np.log10(mfp*ftheta[t]))
     #n1,n2,f1,f2] = self.get_mfp_interpolation(rmfp_log,np.log10(mfp))
      #b0 = kappa * dtheta_ave[t]
    #  b1 = a*ftheta[t]*dtheta[t]
    #  b2 = kappa/mfp/mfp*dtheta_ave[t]
     
      #B1[n1] += f1*b1; B1[n2] += f2*b1
      #B2[n1] += f1*b2; B2[n2] += f2*b2
    #B0 /=np.sum(B0)
    
    
    #B1 /=factor
    #B2 /=np.sum(B2)
    
    #---------------------------------------------
    #Compute DOM
    #----------------
    data.update({'kappa_bulk_tot':sum(kappa_bulk)})
    data.update({'kappa_bulk':kappa_bulk})
    data.update({'mfp_bulk':mfp_bulk})
    data.update({'rmfp':rmfp})
    data.update({'J':J/sum(J)})
    
    data.update({'factor':factor})
    
    #data#.update({'B1':np.tile(B1,(n_rmfp,1))})
    data.update({'B1':B1})

    data.update({'B2':np.tile(B2,(n_rmfp,1))})
    data.update({'rmfp':np.array(rmfp)})

  else: data=None
  self.state = MPI.COMM_WORLD.bcast(data,root=0)


 def get_mfp_interpolation(self,mfp_sampled,mfp_bulk) :

   n_mfp = len(mfp_sampled)
   findm = 0
   for m in range(n_mfp-1) :
    if (mfp_bulk >= mfp_sampled[m]) and (mfp_bulk <= mfp_sampled[m+1]) :
     m1 = m
     m2 = m+1
     mfp1 = mfp_sampled[m1]
     mfp2 = mfp_sampled[m2]
     fm = (mfp_bulk-mfp1)/(mfp2 - mfp1) #for interpolation
     findm = 1
     m_par = [m1,m2,1-fm,fm]
     assert mfp_bulk <= mfp2
     assert mfp_bulk >= mfp1
     assert abs(fm) <= 1
   assert findm == 1
   return m_par




from __future__ import print_function
import os,sys
import numpy as np
import random
import math
#from utils_disorder import *
from shapely.ops import cascaded_union
from shapely.geometry import Point
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
import matplotlib
if not matplotlib.get_backend() == 'Agg': matplotlib.use('Agg')

import networkx as nx
import matplotlib.pylab as plt

#random.seed(3)

def frame_consolidate(argv,polys,frame_tmp):

  d_min =  argv['step']/2.0
  #d_min =  0.0

  frame = Polygon(frame_tmp)
  thin = Polygon(frame_tmp).buffer(1e-4).difference(frame)

  new_p = []
  for nn,p in enumerate(polys):

   inflate = 0
   dd = p.distance(thin)
   if dd < d_min:
    inflate = d_min
    #inflate[nn] = d_min

  #new_p = []
  #for n,poly in enumerate(polys):
   poly = p.buffer(inflate, resolution=1,mitre_limit=20.0,join_style=2)
   new_p.append(poly)


   #new_p2 = []
   #lone pores-----
   #for n in np.where(collision == 0)[0]:
   # new_p2.append(new_p[n])


   #merge pores-----
   #n_coll = 0
   #for n in conn.keys():
   # n_coll +=1
   # tmp = [new_p[n]]
   # for p in conn[n]:
   #  tmp.append(new_p[p])
   # m = MultiPolygon(tmp)
   # new_p2.append(cascaded_union(m))

  #-----------------------
  return new_p








  # coords = p.exterior.coords
  # for n in range(len(coords)-1):
  #  plt.plot([coords[n][0],coords[n+1][0]],[coords[n][1],coords[n+1][1]],color='black')

  # mm = np.mean(coords,axis=0)

  # plt.text(mm[0],mm[1],str(nn))


  #for n in range(4):
  #  n1 = n%4
  #  n2 = (n+1)%4
  #  plt.plot([frame_tmp[n1][0],frame_tmp[n2][0]],[frame_tmp[n1][1],frame_tmp[n2][1]],color='red')


  #plt.axis('equal')
  #plt.show()

  #n = len(polys)
  #newp = []
  #for n1 in range(n):
  # p1 = polys[n1]
  # dd = p1.distance(thin)
  # if dd ==0:
  #  print(n1)
  # if dd < d_min and dd > 0:
  #  print(n1)


  #return newp

def adjust_position(x,y,r,Na,Lx,Ly,d_min):


 poly_clip = []
 dphi = 2.0*math.pi/Na;
 for ka in range(Na):
  ph =  dphi/2 + (ka-1) * dphi
  px  = x + r * math.cos(ph)
  py  = y + r * math.sin(ph)
  poly_clip.append([px,py])
 p = Polygon(poly_clip)


 delta = 1e-4

 tmp = []
 tmp.append([-Lx/2-delta,Ly/2])
 tmp.append([-Lx/2,Ly/2])
 tmp.append([-Lx/2,-Ly/2])
 tmp.append([-Lx/2-delta,-Ly/2])
 L = Polygon(tmp)

 tmp = []
 tmp.append([Lx/2,Ly/2])
 tmp.append([Lx/2+delta,Ly/2])
 tmp.append([Lx/2+delta,-Ly/2])
 tmp.append([Lx/2,-Ly/2])
 R = Polygon(tmp)


 tmp = []
 tmp.append([-Lx/2,Ly/2])
 tmp.append([-Lx/2,Ly/2+delta])
 tmp.append([Lx/2,Ly/2+delta])
 tmp.append([Lx/2,Ly/2])
 U = Polygon(tmp)


 tmp = []
 tmp.append([-Lx/2,-Ly/2])
 tmp.append([-Lx/2,-Ly/2-delta])
 tmp.append([Lx/2,-Ly/2-delta])
 tmp.append([Lx/2,-Ly/2])
 B = Polygon(tmp)


 dx = 0
 dy = 0

 d = p.distance(U)
 if d < d_min and d>0:
  dy += 2*d_min

 d = p.distance(B)
 if d < d_min and d>0:
  dy -= 2*d_min

 d = p.distance(L)
 if d < d_min and d>0:
  dx -= 2*d_min

 d = p.distance(R)
 if d < d_min and d>0:
  dx += 2*d_min


 return dx,dy



def consolidate(polys,argv):

  d_min =  argv['step']/20.0
  #d_min = 0

  #d_min = 0
  n = len(polys)
  conn = {}
  inflate = np.zeros(n)
  collision = np.zeros(n)
  for n1 in range(n):
   p1 = polys[n1]

   for n2 in range(n1+1,n):
    mn =  min([n1,n2])
    Mn =  max([n1,n2])
    p2 = polys[n2]
    if p1.intersects(p2):
     conn.setdefault(mn,[]).append(Mn)
     collision[n1] = 1
     collision[n2] = 1

    dd = p1.distance(p2)
    if dd < d_min and dd > 0:
      inflate[n1] = max([inflate[n1],d_min/2.0])
      inflate[n2] = max([inflate[n2],d_min/2.0])
      conn.setdefault(mn,[]).append(Mn)
      collision[n1] = 1
      collision[n2] = 1


  #Create new pores-----
  new_p = []
  for n,poly in enumerate(polys):
   poly = poly.buffer(inflate[n], resolution=1,mitre_limit=20.0,join_style=2)
   new_p.append(poly)


  new_p2 = []
  #lone pores-----
  for n in np.where(collision == 0)[0]:
   new_p2.append(new_p[n])


  #Cluster conn-------
  G = nx.Graph()
  for n in conn.keys():
   G.add_node(n)
   for m in conn[n]:
    G.add_edge(n,m)
  components = list(nx.connected_components(G))
  #-----------------------

  n_coll = 0
  for c in components:
   n_coll +=1
   tmp = []
   for p in c:
    tmp.append(new_p[p])
   m = MultiPolygon(tmp)
   p = cascaded_union(m)
   new_p2.append(p)


  new_p3 = []
  for p in new_p2:
   pp = p.buffer(d_min, 1, join_style=2).buffer(-d_min, 1, join_style=2)
   new_p3.append(pp)

  #-----------------------
  return new_p3,n_coll




def GenerateRandomPoresOverlap(argv):


  d_min =  argv['step']/20.0
  delta_pore = argv.setdefault('pore_distance',0.2)

  argv.setdefault('shape','square')
  if argv['shape'] == 'circle':
   Na_p = [24]

  if argv['shape'] == 'square':
   Na_p = [4]

  Np = argv['Np']

  random_phi =  argv.setdefault('random_angle',False)

  Lxp =  argv['lx']
  Lyp =  argv['ly']
  Nx = 1
  Ny = 1
  phi_mean =  argv.setdefault('porosity',0.3)
  spread =  argv.setdefault('spread',0.0)



  frame_tmp = []
  frame_tmp.append([float(-Lxp * Nx)/2,float(Lyp * Ny)/2])
  frame_tmp.append([float(Lxp * Nx)/2,float(Lyp * Ny)/2])
  frame_tmp.append([float(Lxp * Nx)/2,float(-Lyp * Ny)/2])
  frame_tmp.append([float(-Lxp * Nx)/2,float(-Lyp * Ny)/2])
  frame = Polygon(frame_tmp)


  Lx = Lxp
  Ly = Lyp
  area_tot = Lx*Ly


  #area_tmp = 0.0;


  #area_max = Lx*phi/float(Np)


  #area_min = Lx*Ly*phi/float(Np)

  #area_bounds = range(2);
  #area_bounds[0] = area_min;
  #area_bounds[1] = area_max;

  #f1 = 1.5
  area_tmp = 0.0;
  cx = []
  cy = []
  Na = []
  radious = []
  angles = []
  phis = []
  pores = [];
  nt_max = 3000
  area_miss = area_tot

  pbc = []
  pbc.append([0,0])
  pbc.append([Lx,0])
  pbc.append([-Lx,0])
  pbc.append([0,Ly])
  pbc.append([0,-Ly])
  pbc.append([Lx,Ly])
  pbc.append([-Lx,-Ly])
  pbc.append([-Lx,Ly])
  pbc.append([Lx,-Ly])
  #---------------------
  #ext = 3.0/4.0
  ext = 0.95
  Xmin = - Lx/2*ext
  Xmax =  Lx/2*ext
  Ymin = - Ly/2*ext
  Ymax =  Ly/2*ext

  final = []
  #pyclipper.SCALING_FACTOR=1e3

  #while (area_miss > 1e-4):
  buf = 0.0
  nn = 0
  confs = []
  #while nn < Np:

  #area = area_min
  Na = Na_p[0]

  if argv.setdefault('manual',False):
   centers = argv['centers']
  else:
   centers = []
   for nn in range(Np):
    x = np.random.uniform(0,1)
    y = np.random.uniform(0,1)
    centers.append([x,y])


  dphi = 2.0*math.pi/Na;
  #Fill polys-----
  polys = []

  for center in centers:
   x = Lx*(center[0]-0.5)
   y = Ly*(center[1]-0.5)
   phi = np.random.uniform(phi_mean - spread, phi_mean + spread)
   area = Lx*Ly*phi/float(Np)
   r = math.sqrt(2.0*area/Na/math.sin(2.0 * math.pi/Na))

   (dx,dy) = adjust_position(x,y,r,Na,Lx,Ly,d_min)

   x +=dx
   y +=dy

   for kp in range(len(pbc)):
    poly_clip = []
    for ka in range(Na):
     ph =  dphi/2 + (ka-1) * dphi
     px  = x + r * math.cos(ph) + pbc[kp][0]
     py  = y + r * math.sin(ph) + pbc[kp][1]
     poly_clip.append([px,py])

    polys.append(Polygon(poly_clip))




  n_coll = 1
  while n_coll > 0:
   polys,n_coll = consolidate(polys,argv)
  #polys,n_coll = consolidate(polys,argv)
  # polys,n_coll = consolidate(polys,argv)
  #polys,n_coll = consolidate(polys,argv)
  #polys,n_coll = consolidate(polys,argv)
  # print(n_coll)


  #n_coll = 1
  #while n_coll > 0:
  # polys,n_coll = consolidate(polys,argv)
  #Create polys
  area = 0.0
  eps = argv['step']/2.0
  polys_cut = []
  for p in polys:
   if p.intersects(frame):
    polys_cut.append(list(p.exterior.coords)[:-1])
    area += p.area

  if argv['save_configuration']:
   #np.array(centers).dump(open('conf.dat','w+'))
   np.array(centers).dump('conf.dat')



  #return frame_tmp,polygons_final
  return frame_tmp,polys_cut



from __future__ import absolute_import
import argparse
from .material import *
from .geometry import *
from .solver import *
from .plot import *

def main(args=None):


 parser = argparse.ArgumentParser()

 #Geometry---------
 parser.add_argument('-g', help='geometry',action='store_true')
 parser.add_argument('-type', help='porous/aligned')
 parser.add_argument('-polyfile', help='file with the coordinates of the polygons')
 parser.add_argument('-porosity', help='<0.25>',default=0.25)
 parser.add_argument('-shape', help='<circle> | square | triangle',default='circle')
 parser.add_argument('-lx', help='<10> [nm]',default=10)
 parser.add_argument('-ly', help='<10> [nm]',default=10)
 parser.add_argument('-step', help='Mesh quality: <1> [nm]',default=1)

 #Material
 parser.add_argument('-m', help='material',action='store_true')
 parser.add_argument('-model', help='<nongray> | gray',default='nongray')
 parser.add_argument('-matfile', help='file with the bulk MFP distribution')
 parser.add_argument('-n_theta', help='Number of azimuthal')
 parser.add_argument('-n_phi', help='Number of polar angles')
 parser.add_argument('-n_mfp', help='Number of mfps')

 #Solver
 parser.add_argument('-s', help='solver',action='store_true')
 parser.add_argument('-multiscale', help='solver',action='store_true')
 parser.add_argument('-max_bte_iter', help='solver',default=10)

 #Plot
 parser.add_argument('-p', help='plot',action='store_true')
 parser.add_argument('-variable', help='Plot Results: map/temperature')
 parser.add_argument('-repeat_x',default=3,help='Repetition along x')
 parser.add_argument('-repeat_y',default=3,help='Repetition along y')

 args = parser.parse_args()

 #Create Geometry----------------------------------------

 if args.g:
  Geometry(**vars(args))

 if args.m:
  Material(**vars(args))

 if args.s:
  Solver(**vars(args))

 if args.p:
  Plot(**vars(args))

if __name__ == "__main__":

    main()



import numpy as np
import os
import deepdish as dd
from mpi4py import MPI
from numpy.testing import assert_array_equal


class Material(object):

 def __init__(self,**argv):

  if MPI.COMM_WORLD.Get_rank() == 0:
  
   output = {}   
   data = self.compute_mat_2D(argv)   
   output.update({'data_2D':data})
   
   data = self.compute_mat_3D(argv)   
   output.update({'data_3D':data})
   
   dd.io.save('material.hdf5',output)   
  MPI.COMM_WORLD.Barrier()

 def compute_mat_2D(self,argv):  
   
  if MPI.COMM_WORLD.Get_rank() == 0:
   #Polar Angle-----
   n_phi = int(argv.setdefault('n_phi',46)); Dphi = 2*np.pi/n_phi
   phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
   #--------------------
   
   #Azimuthal Angle------------------------------
   n_theta = int(argv.setdefault('n_theta',48)); Dtheta = np.pi/n_theta/2.0
   theta = np.linspace(Dtheta/2.0,np.pi/2.0 - Dtheta/2.0,n_theta)
   dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)   
   
   #Compute directions---
   polar = np.array([np.sin(phi),np.cos(phi),np.ones(n_phi)]).T
   azimuthal = np.array([np.sin(theta),np.sin(theta),np.cos(theta)]).T
   direction = np.einsum('ij,kj->ikj',azimuthal,polar)
   
   #Compute average---
   ftheta = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
   fphi= np.sinc(Dphi/2.0/np.pi)
   polar_ave = np.array([fphi*np.ones(n_phi),fphi*np.ones(n_phi),np.ones(n_phi)]).T
   azimuthal_ave = np.array([ftheta,ftheta,np.cos(Dtheta/2)*np.ones(n_theta)]).T   
   direction_ave = np.multiply(np.einsum('ij,kj->ikj',azimuthal_ave,polar_ave),direction)
   #---------------------------------------------------------------
   
   #Import material
   if argv.setdefault('model','nongray')=='nonfray':
    tmp = np.loadtxt(os.path.dirname(__file__) + '/materials/'+ argv['matfile'])
    mfp_bulk = tmp[:,0]*1e9; 
    kappa_bulk = tmp[:,1]
   else:
    mfp_bulk = np.array(argv['mfp'],dtype='double'); 
    kappa_bulk = argv.setdefault('kappa_bulk',np.ones(len(mfp_bulk)))

   n_mfp_bulk = len(mfp_bulk) 
   mfp = np.logspace(-3,np.log10(max(mfp_bulk)),argv.setdefault('n_mfp',100)) 
   trials = np.outer(mfp_bulk,ftheta*np.sin(theta))
   J0 = np.outer(np.ones(n_mfp_bulk),dtheta)
   J1 = np.outer(kappa_bulk/mfp_bulk,ftheta*dtheta*np.sin(theta))/sum(kappa_bulk/mfp_bulk)   
   J2 = np.outer(kappa_bulk/pow(mfp_bulk,2),dtheta)/sum(kappa_bulk/pow(mfp_bulk,2))   #x2 (symmetry) and /2 (average)
   J3 = np.outer(1/mfp_bulk,ftheta*dtheta*np.sin(theta))

   G = mfp/2
   kappa_mfe = np.square(mfp)/2

   domega = np.ones(n_phi)*Dphi/2.0/np.pi
 
   data = {'kappa_bulk_tot':sum(kappa_bulk),'kappa_bulk':kappa_bulk,\
                'mfp_bulk':mfp_bulk,\
                'fphi':fphi,\
                'G':G,\
                'kappa_mfe':kappa_mfe,\
                'dphi':Dphi,\
                'n_mfp':len(mfp),\
                'trials':trials,\
                'domega':domega,\
                'mfp':mfp,\
                'J0':J0,'J1':J1,'J2':J2,'J3':J3,\
                'n_phi':n_phi,'n_theta':n_theta,\
                'direction':direction,\
                'direction_ave':direction_ave,\
                'polar':polar,\
                'azimuthal':azimuthal,\
                'polar_ave':polar,\
                'control_angle':polar,\
                'azimuthal_ave':azimuthal}
   
   return data

 def compute_mat_3D(self,argv):  
   
  if MPI.COMM_WORLD.Get_rank() == 0:
    #Polar Angle-----
    n_phi = int(argv.setdefault('n_phi',46)); Dphi = 2.0*np.pi/n_phi
    phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
   #--------------------
   
    #Azimuthal Angle------------------------------
    n_theta = int(argv.setdefault('n_theta',48)); Dtheta = np.pi/n_theta
    theta = np.linspace(Dtheta/2.0,np.pi - Dtheta/2.0,n_theta)
    dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)   
   
    #Compute directions---
    polar = np.array([np.sin(phi),np.cos(phi),np.ones(n_phi)]).T
    azimuthal = np.array([np.sin(theta),np.sin(theta),np.cos(theta)]).T
    direction = np.einsum('ij,kj->ikj',azimuthal,polar)
   
    #Compute average---
    ftheta = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
    fphi= np.sinc(Dphi/2.0/np.pi)
    polar_ave = np.array([fphi*np.ones(n_phi),fphi*np.ones(n_phi),np.ones(n_phi)]).T
    azimuthal_ave = np.array([ftheta,ftheta,np.cos(Dtheta/2)*np.ones(n_theta)]).T   
    direction_ave = np.multiply(np.einsum('ij,kj->ikj',azimuthal_ave,polar_ave),direction)
    #---------------------------------------------------------------
    control_angle = direction_ave.reshape((n_phi*n_theta,3))

    
    #Import material
    if argv.setdefault('model','nongray')=='nonfray':
     tmp = np.loadtxt(os.path.dirname(__file__) + '/materials/'+ argv['matfile'])
     mfp_bulk = tmp[:,0]*1e9; n_mfp_bulk = len(mfp_bulk)
     kappa_bulk = tmp[:,1]
    else: 
     mfp_bulk = np.array(argv['mfp'],dtype='double'); 
     kappa_bulk = argv.setdefault('kappa_bulk',np.ones(len(mfp_bulk)))

    n_mfp_bulk = len(mfp_bulk)
    J0 = np.ones(n_mfp_bulk)
    J1 = kappa_bulk/mfp_bulk/sum(kappa_bulk/mfp_bulk)   
    J2 = kappa_bulk/pow(mfp_bulk,2)/sum(kappa_bulk/pow(mfp_bulk,2))   
    J3 = 1/mfp_bulk

    
    mfp = np.logspace(-3,np.log10(max(mfp_bulk)),argv.setdefault('n_mfp',100))
    trials = mfp_bulk
    J0 = np.array([J0]).T
    J1 = np.array([J1]).T
    J2 = np.array([J2]).T
    J3 = np.array([J3]).T
    trials = np.array([trials]).T
    
    G = mfp/2
    kappa_mfe = np.square(mfp)/3

   
    domega = []
    for t in range(n_theta):
     for p in range(n_phi):
      domega.append(dtheta[t]*Dphi/4.0/np.pi)
    domega = np.array(domega)
   


    data = {'kappa_bulk_tot':sum(kappa_bulk),'kappa_bulk':kappa_bulk,\
                'mfp_bulk':mfp_bulk,\
                'fphi':fphi,\
                'dphi':Dphi,\
                'n_mfp':len(mfp),\
                'G':G,\
                'kappa_mfe':kappa_mfe,\
                'trials':trials,\
                'domega':domega,\
                'mfp':mfp,\
                'J0':J0,'J1':J1,'J2':J2,'J3':J3,\
                'n_phi':n_phi,'n_theta':n_theta,\
                'direction':direction,\
                'control_angle':control_angle,\
                'direction_ave':direction_ave,\
                'polar':polar,\
                'azimuthal':azimuthal,\
                'polar_ave':polar,\
                'azimuthal_ave':azimuthal}
   
    return data
   

   



from __future__ import absolute_import
from mpi4py import MPI
from .solver_new import Solver
from pyvtk import *
import numpy as np
import deepdish as dd
import os
import matplotlib
if not matplotlib.get_backend() == 'Agg': matplotlib.use('Agg')
import matplotlib.pylab as plt
import matplotlib.patches as patches
from matplotlib.path import Path
from .fig_maker import *
from shapely import geometry,wkt
from scipy.interpolate import BarycentricInterpolator
from .WriteVTK import *
from .geometry import *
from scipy.interpolate import griddata
from shapely.geometry import MultiPoint,Point,Polygon,LineString
import shapely

def get_suppression(mfps,sup,mfp):


  if mfp < mfps[0]:
   output = sup[0]
  else:
   included = False
   for n in range(len(mfps)-1):
    mfp_1 = mfps[n]
    mfp_2 = mfps[n+1]
    if mfp < mfp_2 and mfp > mfp_1:
     output = sup[n] + (sup[n+1]-sup[n])/(np.log10(mfps[n+1])-np.log10(mfps[n]))*(np.log10(mfp)-np.log10(mfps[n]))
     included = True
     break

   if included == False:
    output = sup[-1]*mfps[-1]/mfp

  return output


def create_path(obj):

   codes = [Path.MOVETO]
   for n in range(len(obj)-1):
    codes.append(Path.LINETO)
   codes.append(Path.CLOSEPOLY)

   verts = []
   for tmp in obj:
     verts.append(tmp)
   verts.append(verts[0])

   path = Path(verts,codes)
   return path

class Plot(object):

 def __init__(self,**argv):


  if MPI.COMM_WORLD.Get_rank() == 0:
   self.solver = dd.io.load('solver.hdf5')
   self.Nx = argv.setdefault('repeat_x',1)
   self.Ny = argv.setdefault('repeat_y',1)

   if '/' in argv['variable']:
    model= argv['variable'].split('/')[0]
    if model == 'map':
     self.plot_map(argv)
    if model == 'line':
     self.plot_over_line(argv)

   #if argv['variable'] == 'suppression_function' :
   # self.plot_suppression_function(argv)

   if argv['variable'] == 'SUP' :
    self.SUP()


   if argv['variable'] == 'distribution' :
    self.plot_distribution(argv)

   if argv['variable'] == 'vtk' :
    self.save_vtk(argv)


   if argv['variable'] == 'kappa_bte' :
    solver = dd.io.load('state.hdf5')
    kappa_bte = solver['kappa_bte']
    if argv.setdefault('save',True) :
     f = open('kappa_bte.dat','w+')
     f.write(str(kappa_bte))
     f.close()

 def save_vtk(self,argv):
  #if MPI.COMM_WORLD.Get_rank() == 0:
   #Write data-------
   #geo = dd.io.load('geometry.hdf5')
   geo = Geometry(type='load')
   solver = dd.io.load('solver.hdf5')
   argv.update({'Geometry':geo})
   vw = WriteVtk(argv)
   #vw.add_variable(solver['fourier_temperature'],label = 'Fourier Temperature [K]')
   #vw.add_variable(solver['fourier_flux'],label = r'''Thermal Flux [W/m/m]''')

   vw.add_variable(solver['temperature'],label = r'''BTE Temperature [K]''')
   vw.add_variable(solver['flux'],label = r'''BTE Thermal Flux [W/m/m]''')
   vw.add_variable(solver['temperature_fourier'],label = r'''Fourier Temperature [K]''')
   vw.add_variable(solver['flux_fourier'],label = r'''Fourier Thermal Flux [W/m/m]''')


   vw.write_vtk()


 def SUP(self):

    SUP = self.solver['SUP'] 
    MFP = self.solver['MFP']

    plt.plot(MFP,SUP)

    plt.xscale('log')

    plt.show()


 def plot_material(self,argv):
  if MPI.COMM_WORLD.Get_rank() == 0:
   init_plotting()
   data = dd.io.load('material.hdf5')
   #dis_nano = list(data['B0']*data['kappa_bulk_tot'])
   dis_bulk = list(data['kappa_bulk'])
   mfp = list(np.multiply(data['mfp_bulk'],1e6))
   #mfp_nano = list(np.multiply(data['mfp_sampled'],1e6))
   fill([mfp[0]] + mfp + [mfp[-1]],[0] + dis_bulk + [0])

   #fill([mfp_nano[0]] + mfp_nano + [mfp_nano[-1]],[0] + dis_nano + [0],color=c2,alpha=0.3)

   xscale('log')
   #grid(which='both')
   xlabel('$\Lambda_{\mathrm{b}}$ [$\mu$ m]')
   ylabel('$K_{\mathrm{b}}d\Lambda_{\mathrm{b}}$ [Wm$^{-1}$K$^{-1}$]')
   ylim([0,max(dis_bulk)*1.3])
   show()



 def plot_distribution(self,argv):
  if MPI.COMM_WORLD.Get_rank() == 0:
   init_plotting(extra_x_padding = 0.05)
   data = dd.io.load('material.hdf5')
   dis_bulk = list(data['kappa_bulk'])
   mfp = list(data['mfp_bulk']) #m

   #plot nano------
   solver = dd.io.load(argv.setdefault('filename','solver.hdf5'))
   mfp_sampled = data['mfp_sampled'] #m
   suppression = solver['suppression']
   sup = [suppression[m,:,:].sum() for m in range(np.shape(suppression)[0])]
   kappa_nano = []
   mfp_nano = []
   for n in range(len(mfp)):
    sup_interp = get_suppression(mfp_sampled,sup,mfp[n])
    kappa_nano.append(dis_bulk[n]*sup_interp)
    mfp_nano.append(mfp[n]*sup_interp*1e9)

   #reordering-----
   I = np.argsort(np.array(mfp_nano))
   kappa_sorted = []
   for i in I:
    kappa_sorted.append(kappa_nano[i])
   mfp_nano = sorted(mfp_nano)

   #----------------
   mfp_bulk = np.array(mfp)*1e9


   #fill([mfp_bulk[0]] + list(mfp_bulk) + [mfp_bulk[-1]+1e-6],[0] + dis_bulk + [0],lw=2,alpha=0.6)
   fill([mfp_nano[0]] + mfp_nano + [mfp_nano[-1]],[0] + kappa_sorted + [0],color=c2)

   f = open('mfp_nano.dat','w+')
   for n in range(len(mfp)):
    f.write('{0:10.4E} {1:20.4E}'.format(mfp_nano[n]*1e-9,kappa_sorted[n]))
    f.write('\n')
   f.close()

   xscale('log')
   xlabel('$\Lambda$ [nm]')
   ylabel('$Kd\Lambda$ [Wm$^{-1}$K$^{-1}$]')
   grid()
   show()


 def plot_over_line(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:
   self.geo = Geometry(type='load')
   (data,nodes,triangulation) = self.get_data(argv)
   (Lx,Ly) = self.geo.get_repeated_size(argv)

   #--------
   p1 = np.array([-Lx/2.0+1e-7,0.0])
   p2 = np.array([Lx/2.0*(1-1e-7),0.0])
   (x,data,int_points) = self.compute_line_data(p1,p2,data)
   #--------
   axes([0.5,0.01,0.5,1.0])
   plot(x,data,linewidth=2)

   for p in int_points:
    plot([p,p],[min(data),max(data)],ls='--',color='k')
   plot([min(x),max(x)],[0,0],ls='--',color='k')
   gca().yaxis.tick_right()


 def point_in_elem(self,elem,p):
  poly = []
  for n in self.geo.elems[elem]:
   t1 = self.geo.nodes[n][0]
   t2 = self.geo.nodes[n][1]
   poly.append((t1,t2))
  polygon = Polygon(poly)
  if polygon.contains(Point(p[0],p[1])):
   return True
  else:
   return False

 def compute_line_data(self,p1,p2,original):

  N = 1000
  tt = []
  delta = np.linalg.norm(p2-p1)/N
  neighbors = range(len(self.geo.elems))
  h1 = 0
  h2 = 0
  gamma = 1e-4
  data = []
  x = []
  int_points = []
  for n in range(N+1):
   r = p1 + n*delta*(p2-p1)
   #Here we include extra points close to the interfaces-------
   if n > 0:
    tmp = self.geo.cross_interface(p1 + (n-1)*delta*(p2-p1),r)
    versor = r - p1 + (n-1)*delta*(p2-p1)
    versor /= np.linalg.norm(versor)
    if len(tmp) > 0:
     pp = [tmp-gamma*versor,tmp + gamma*versor,r]
     x.append(n*delta-np.linalg.norm(r-tmp)-gamma)
     x.append(n*delta-np.linalg.norm(r-tmp)+gamma)
     x.append(n*delta)
     int_points.append(n*delta-np.linalg.norm(r-tmp)-gamma)
    else:
     pp = [r]
     x.append(n*delta)
   else:
    pp = [r]
    x.append(n*delta)
   #----------------------------------------------------------
   for p in pp:
    is_in = False
    for elem in neighbors:
     if self.point_in_elem(elem,p) :
      is_in = True
      value = self.geo.compute_2D_interpolation(original,p,elem)
      data.append(value)
      neighbors = self.geo.get_elem_extended_neighbors(elem)
      break
    if not is_in:
     for elem in range(len(self.geo.elems)):
      if self.point_in_elem(elem,p) :
       neighbors = self.geo.get_elem_extended_neighbors(elem)
       value = self.geo.compute_2D_interpolation(original,p,elem)
       data.append(value)
       break

  return x,data,int_points


 def get_data(self,argv):


  size = self.geo.size
  Nx = argv.setdefault('repeat_x',1)
  Ny = argv.setdefault('repeat_y',1)
  Nz = argv.setdefault('repeat_z',1)
  Lx = size[0]*Nx
  Ly = size[1]*Ny
  variable = argv['variable'].split('/')[1]
  solver = dd.io.load('solver.hdf5')
  argv.update({'Geometry':self.geo})

  vw = WriteVtk(argv)
  (triangulation,tmp,nodes) = vw.get_node_data(solver[variable])



  if argv['direction'] == 'x':
     data = np.array(tmp).T[0]
  elif argv['direction'] == 'y':
     data = np.array(tmp).T[1]
  elif argv['direction'] == 'z':
     data = np.array(tmp).T[2]
  elif argv['direction'] == 'magnitude':
      data = []
      for d in tmp:
       data.append(np.linalg.norm(d))
      data = np.array(data)
  else:
    data = np.array(tmp).T[0]

  return data,nodes,triangulation


 def plot_map(self,argv):


   self.geo = Geometry(type='load')
   (Lx,Ly) = self.geo.get_repeated_size(argv)

   fig = figure(num=' ', figsize=(8*Lx/Ly, 8), dpi=80, facecolor='w', edgecolor='k')
   #axes([0,0,0.5,1.0])
   axes([0,0,1.0,1.0])


   #data += 1.0
   (data,nodes,triangulation) =  self.get_data(argv)
   vmin = argv.setdefault('vmin',min(data))
   vmax = argv.setdefault('vmax',max(data))

   tripcolor(triangulation,np.array(data),shading='gouraud',norm=mpl.colors.Normalize(vmin=vmin,vmax=vmax),zorder=1)


   #colorbar(norm=mpl.colors.Normalize(vmin=min(data),vmax=max(data)))

    #Contour-----
   if argv.setdefault('iso_values',True):
    t = tricontour(triangulation,data,levels=np.linspace(vmin,vmax,10),colors='black',linewidths=1.5)



   if argv.setdefault('streamlines',False):# and (variable == 'fourier_flux' or variable == 'flux' ):


       n_lines = argv.setdefault('n_lines',10)*self.Ny #assuming transport across x
       xi = np.linspace(-Lx*0.5,Lx*0.5,300*self.Nx)
       yi = np.linspace(-Ly*0.5,Ly*0.5,300*self.Ny)
       x = np.array(nodes)[:,0]
       y = np.array(nodes)[:,1]
       z = np.array(data).T[0]
       Fx = griddata((x, y), z, (xi[None,:], yi[:,None]), method='cubic')
       z = np.array(data).T[1]
       Fy = griddata((x, y), z, (xi[None,:], yi[:,None]), method='cubic')


       seed_points = np.array([-Lx*0.5*0.99*np.ones(n_lines),np.linspace(-Ly*0.5*0.98,Ly*0.5*0.98,n_lines)])
       ss = streamplot(xi, yi, Fx, Fy,maxlength = 1e8,start_points=seed_points.T,integration_direction='both',color='r',minlength=0.95,linewidth=1)


   if argv.setdefault('plot_interfaces',False):
    pp = self.geo.get_interface_point_couples(argv)


   axis('off')
   gca().invert_yaxis()
   xlim([-Lx*0.5,Lx*0.5])
   ylim([-Ly*0.5,Ly*0.5])
   axis('equal')
   show()


 def plot_suppression_function(self,argv):

  #if MPI.COMM_WORLD.Get_rank() == 0:
   #init_plotting(extra_bottom_padding= 0.01,extra_x_padding = 0.02)
   data = dd.io.load(argv.setdefault('filename','solver.hdf5'))

   suppression = data['suppression']
   iso_sup = data['iso_suppression']
   #iso_sup = [iso_suppression[m].sum() for m in range(np.shape(iso_suppression)[0])]

   #print(suppression)
   #quit()
   tmp = data['fourier_suppression']
   sup = [suppression[m,:,:].sum() for m in range(np.shape(suppression)[0])]
   sup_fourier = [tmp[m,:,:].sum() for m in range(np.shape(tmp)[0])]


   mfp = data['mfp']*1e6
   #print(sup_fourier[0])
   if argv.setdefault('show',True):
    plot(mfp,sup,color=c1)
    plot(mfp,sup_fourier,color=c2)
    plot(mfp,iso_sup,color=c3)

    xscale('log')
    grid('on')
    xlabel('Mean Free Path [$\mu$ m]')
    ylabel('Suppression Function')
    show()
   if argv.setdefault('write',True):
    suppression.dump('suppression.dat')



import numpy as np

nodes = {}; elem_sides_map = {}; side_nodes_map = {}
node_elems_map = {}; node_sides_map = {}; elem_nodes_map = {}; side_elems_map = {}

with open('mesh.msh') as f:
 [f.readline() for n in range(13)] 
 nodes = {n:np.array(f.readline().split()[1:4],'float') for n in range(int(f.readline()))}
 [f.readline() for n in range(2)] 
 for i in range(int(f.readline())):
  e = np.array(f.readline().split()[5:8],'int')
  [node_elems_map.setdefault(p,set()).add(i) for p in e] 
  [elem_nodes_map.setdefault(i,set()).add(p) for p in e] 
  elem = []   
  for n in range(3):   
   tmp = sorted(np.roll(e,n)[0:2])
   if not tmp in side_nodes_map.values():
    side_nodes_map.update({len(side_nodes_map):tmp})   
    for p in tmp: node_sides_map.setdefault(p,set()).add(len(side_nodes_map))
    ns = len(side_nodes_map) 
   else:
    ns = list(side_nodes_map.keys())[list(side_nodes_map.values()).index(tmp)]
   elem_sides_map.setdefault(i,set()).add(ns)
   side_elems_map.setdefault(ns,set()).add(i)




    







from __future__ import print_function
#import xml.etree.ElementTree as ET
#import os
#import sys
import numpy as np
#import random
#import math
import matplotlib.pylab as plt
#import subprocess
from shapely.ops import cascaded_union
from shapely.geometry import Point,LineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
import shapely

#def get_loop(points,lines):


# return loop



def line_exists_ordered(l,lines):
 for n,line in enumerate(lines) :
  if (line[0] == l[0] and line[1] == l[1]) :
   return n
  if (line[0] == l[1] and line[1] == l[0]) :
   return -n
 return 0


def line_exists(l,lines):

 for n,line in enumerate(lines) :
   if (line[1] == l[0] and line[2] == l[1]) or (line[1] == l[1] and line[2] == l[0]) :
     return line
 return 0

def point_exists(p,points):

 delta = 1e-10
 for n,point in enumerate(points) :
   if abs(point[0] - p[0]) < delta and abs(point[1] - p[1])< delta :
     return n

 return -1


def get_line_from_points(p1,p2,all_lines,points):

 #Get indexes
 for t,p in enumerate(points):
  if p == p1 :
   t1 = t
  if p == p2 :
   t2 = t


 for t,l in enumerate(all_lines):
  if (t1 == l[1] and t2 == l[2]) :
   line = [l[0],t1,t2]
   break
  if (t2 == l[1] and t1 == l[2])  :
   line = [l[0],t2,t1]
   break


 return line


#def rev_rotate(l,n):
#    return l[-n:] + l[:-n]


#def rotate(l,n):
#    return l[n:] + l[:n]


#def GetSampleList(output_dir) :
#   tmp =  glob.glob(output_dir + '/*.json')
#   index = 0
#   if len(tmp) > 0:
#    for n in range(len(tmp)) :
#     tmp2 = tmp[n].split('/')
#     tmp3 = tmp2[-1].split('.')
#     tmp4 = tmp3[0].split('_')
#     index = max([int(tmp4[1]),index])
#    index +=1


#   return index



def plot_region(region):

   for n in range(len(region)) :
    n1 = n
    n2 = (n+1)%len(region)
    plt.plot([region[n1][0],region[n2][0]],[region[n1][1],region[n2][1]],color='black')
   plt.xlim([-3,3])
   plt.ylim([-3,3])


def FindPeriodic(control,others,points,lines):

 #line1 = LineString([points[lines[control][0]],points[lines[control][1]]])  
    

 delta = 1e-3
 p1c = np.array(points[lines[control][0]])
 p2c = np.array(points[lines[control][1]])
 pc = (p1c + p2c)/2.0
 
     
 for n in range(len(others)):
  p1o = np.array(points[lines[others[n]][0]])
  p2o = np.array(points[lines[others[n]][1]])
  po = (p1o + p2o)/2.0   
  
  
  if np.linalg.norm(np.dot(p1c-p2c,pc-po)) < delta:     
   #print(p1c,p2c)
   #print(p1o,p2o)
   line1 = LineString([p1c,p1o])
   line2 = LineString([p2c,p2o])     
   if line1.intersects(line2):
    return -others[n]
   else:
    return others[n]
       
  #print(p1c,p2c)
   #line1 = LineString([points[lines[others[n]][0]],points[lines[others[n]][1]]])  
   #if line1.distance(line2)

   #Positive
   #pd1 = [p1c[0]-p1[0],p1c[1]-p1[1]]
   #pd2 = [p2c[0]-p2[0],p2c[1]-p2[1]]
   #if abs(pd1[0] - pd2[0]) < delta and abs(pd1[1] - pd2[1]) <delta :
   #  return others[n]

   #Negative
   #pd1 = [p1c[0]-p2[0],p1c[1]-p2[1]]
   #pd2 = [p2c[0]-p1[0],p2c[1]-p1[1]]
   #if abs(pd1[0] - pd2[0])<delta and abs(pd1[1] - pd2[1])<delta :
   #  return -others[n]


 #If no periodic line has been found
 print("NO PERIODIC LINES HAS BEEN FOUND")
 quit()

def LineOrdering(lines) :

  new_lines = []
  new_lines.append(lines[0])
  check = np.ones(len(lines))
  check[0] = 0
  n_check = 0
  while len(lines) > len(new_lines) :
   n_check +=1
   line = new_lines[len( new_lines)-1]
   for k in range(len(lines)):
    if check[k] == 1:
     if lines[k][1] == line[2]  :
      check[k] = 0
      new_lines.append([lines[k][0],lines[k][1],lines[k][2]])
      break;
     if lines[k][2] == line[2]  :
      check[k] = 0
      new_lines.append([-lines[k][0],lines[k][2],lines[k][1]])
      break;
   if n_check > len(lines) :
    print('CANNOT FIND A ORDERED LOOP')
    print(lines)
    quit()
  return new_lines



def PolygonArea(corners):
    n = len(corners) # of corners
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += corners[i][0] * corners[j][1]
        area -= corners[j][0] * corners[i][1]
    area = abs(area) / 2.0
    return area


def already_included(all_points,new_point):

 for n,p in enumerate(all_points):
  d = np.linalg.norm(np.array(p)-np.array(new_point))
  if d < 1e-12:
   return n
 return -1


def regularize_polygons(polygons):


 #cut redundant points
 new_poly = []
 for poly in polygons:
  N = len(poly)
  tmp = []
  for n in range(N):
   p1 = poly[n]
   p2 = poly[(n+1)%N]
   if np.linalg.norm(np.array(p1)-np.array(p2)) >1e-4:
    tmp.append(p1)
  new_poly.append(tmp)


 #eliminate useless points
 #cut redundant points
 #new_poly2 = []
 #for poly in new_poly:
 # poly.append(poly[-1])
 # N = len(poly)
 # tmp = []
 # for n in range(N):
 #  p1 = np.array(poly[(n-1)%N])
 #  p2 = np.array(poly[n])
 #  p3 = np.array(poly[(n+1)%N])

 #  l1 = (p2-p1)/np.linalg.norm(p2-p1)
 #  l2 = (p3-p2)/np.linalg.norm(p3-p2)

 #  theta = np.arccos(np.dot(l1,l2))
 #  tmp.append(p1)
#   else:
#    print('ggg')
 # new_poly2.append(tmp)


 return new_poly

def PolyArea(x,y):
    return 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))

def prune(solution):


 #Eliminate unnecessary points---
 tmp = []
 for region in solution:
  points = []
  for p,point in enumerate(region):
   if p > 0:
    if already_included([points[-1]],point) == -1:
     points.append(point)
   else:
    points.append(point)
  tmp.append(points)


 polygons = []
 for n,region in enumerate(tmp):

  area = PolyArea(np.array(region)[:,0],np.array(region)[:,1])
  if area > 1e-2:
   points = []

   for p,new_point in enumerate(region):
    tmp = already_included(points,new_point)
    if not tmp == -1 and not tmp == len(points)-1:
     polygons.append(points[tmp:len(points)])
     del points[tmp+1:len(points)]
    if not (not tmp == -1 and tmp == len(points)-1):
     points.append(new_point)
   polygons.append(points)


 #for poly in polygons:
 # poly.append(poly[0])
 # x = np.array(poly)[:,0]
 # y = np.array(poly)[:,1]
 # plot(x,y)
 # show()

 return polygons


def compute_line_point_distance(p1,p2,p3):
   return np.linalg.norm(np.cross(p2-p1, p3-p1))/np.linalg.norm(p2-p1)

def plot_region(dd,color='r'):

 dd.append(dd[0])
 dd = np.array(dd)
 x = dd[:,0]
 y = dd[:,1]
 plot(x,y,color=color)
 axis('equal')
 axis('off')
 xlim([-30,30])
 ylim([-30,30])


def create_line_list(pp,points,lines,store,mesh_ext):


   p_list = []
   for p in pp:
    tmp = already_included(points,p)
    if tmp == -1:
     points.append(p)
     store.write( 'Point('+str(len(points)-1) +') = {' + str(p[0]) +','+ str(p[1])+',0,'+ str(mesh_ext) +'};\n')
     p_list.append(len(points)-1)
    else:
     p_list.append(tmp)


   line_list = []
   for l in range(len(p_list)):
    p1 = p_list[l]
    p2 = p_list[(l+1)%len(p_list)]
    if not p1 == p2:
    #f 1 == 1:    
     tmp = line_exists_ordered([p1,p2],lines)

     if tmp == 0 : #craete line
      lines.append([p1,p2])
      store.write( 'Line('+str(len(lines)-1) +') = {' + str(p1) +','+ str(p2)+'};\n')
      line_list.append(len(lines)-1)
     else:
      line_list.append(tmp)

   
   return line_list


def create_loop(loops,line_list,store):

   #create external loop
   strc = 'Line Loop(' + str(loops)+ ') = {'
   for n in range(len(line_list)):
    strc +=str(line_list[n])
    if n == len(line_list)-1:
     strc += '};\n'
    else:
     strc += ','
   store.write(strc)

   return strc



def create_surface(ss,bulk_surface,store):


  strc = '''Plane Surface(''' + str(bulk_surface) + ')= {'
  for n,s in enumerate(ss):
   strc += str(s)
   if n == len(ss)-1:
     strc += '};\n'
   else:
     strc += ','
  store.write(strc)


def mesh(polygons,frame,argv):


  polygons = regularize_polygons(polygons)

  mesh_ext = argv['step']
  
  Frame = Polygon(frame)

  #CREATE BULK SURFACE------------------------------------
  store = open('mesh.geo', 'w+')
  points = []
  lines = []
  loops = 0
  ss = 0
  polypores = []


  for poly in polygons:
   thin = Polygon(poly).intersection(Frame)
   if isinstance(thin, shapely.geometry.multipolygon.MultiPolygon):
     tmp = list(thin)
     for t in tmp:
      polypores.append(t)
   else:
    polypores.append(thin)

   #Points-------
   #pp = list(thin.exterior.coords)[:-1]

  #Get boundary wall
  lx = frame[0][1]*2.0
  ly = frame[1][1]*2.0
  pore_wall = []
  delta = 1e-2

  #-------------------------

  #Create bulk surfacep
  MP = MultiPolygon(polypores)
  bulk = Frame.difference(cascaded_union(MP))
  if not (isinstance(bulk, shapely.geometry.multipolygon.MultiPolygon)):
   bulk = [bulk]

  bulk_surface = []
  inclusions = []
  for region in bulk:

   pp = list(region.exterior.coords)[:-1]
   line_list = create_line_list(pp,points,lines,store,mesh_ext)

   loops +=1
   local_loops = [loops]
   create_loop(loops,line_list,store)

   #Create internal loops-------------
   for interior in region.interiors:
    pp = list(interior.coords)[:-1]
    line_list = create_line_list(pp,points,lines,store,mesh_ext)
    loops +=1
    local_loops.append(loops)
    create_loop(loops,line_list,store)
    #if argv.setdefault('inclusion',False):
    # ss +=1
    # inclusions.append(ss)
    # create_surface([loops],ss,store)
   #---------------------------------

   ss +=1
   bulk_surface.append(ss)
   create_surface(local_loops,ss,store)

  #Create Inclusion surfaces--------
  if argv.setdefault('inclusion',False):

   for poly in polygons:
    thin = Polygon(poly).intersection(Frame)

    #Points-------
    pp = list(thin.exterior.coords)[:-1]
    line_list = create_line_list(pp,points,lines,store,mesh_ext)
    loops +=1
    create_loop(loops,line_list,store)
    ss +=1
    create_surface([loops],ss,store)
    inclusions.append(ss)

   strc = r'''Physical Surface('Inclusion') = {'''
   for n,s in enumerate(inclusions):
    strc += str(s)
    if n == len(inclusions)-1:
     strc += '};\n'
    else:
      strc += ','
   store.write(strc)
  #-------------------------------


  strc = r'''Physical Surface('Matrix') = {'''
  for n,s in enumerate(bulk_surface):
   strc += str(s)
   if n == len(bulk_surface)-1:
     strc += '};\n'
   else:
     strc += ','
  store.write(strc)


  hot = []
  cold = []
  upper = []
  lower = []

  #deltax = (Maxx-Minx)/1e2
  #deltay = (Maxy-Miny)/1e2

  pul = np.array([-lx/2.0,ly/2.0])
  pur = np.array([lx/2.0,ly/2.0])
  pll = np.array([-lx/2.0,-ly/2.0])
  plr = np.array([lx/2.0,-ly/2.0])


  delta = 1e-8
  pore_wall = []
  for l,line in enumerate(lines):
  
   pl = (np.array(points[line[0]])+np.array(points[line[1]]))/2.0
   
   is_on_boundary = True
   if compute_line_point_distance(pul,pur,pl) < delta:     
     upper.append(l)
    
     is_on_boundary = False
     
   if compute_line_point_distance(pll,plr,pl) < delta:
     lower.append(l)
     is_on_boundary = False 
     
   if compute_line_point_distance(plr,pur,pl) < delta:
     cold.append(l)
     is_on_boundary = False 
     
   if compute_line_point_distance(pll,pul,pl) < delta:
     hot.append(l)
     is_on_boundary = False   
     
   
   if is_on_boundary:
    
    pore_wall.append(l)


  #quit()
  additional_boundary = []
  if argv.setdefault('Periodic',[True,True,True])[0]:
   strc = r'''Physical Line('Periodic_1') = {'''
   for n in range(len(hot)-1):
    strc += str(hot[n]) + ','
   strc += str(hot[-1]) + '};\n'
   store.write(strc)

   strc = r'''Physical Line('Periodic_2') = {'''
   for n in range(len(cold)-1):
    strc += str(cold[n]) + ','
   strc += str(cold[-1]) + '};\n'
   store.write(strc)
  else:
   for k in hot:
    additional_boundary.append(k)
   for k in cold:
    additional_boundary.append(k)

  if argv.setdefault('Periodic',[True,True,True])[1]:
   strc = r'''Physical Line('Periodic_3') = {'''
   for n in range(len(upper)-1):
    strc += str(upper[n]) + ','
   strc += str(upper[-1]) + '};\n'
   store.write(strc)

   strc = r'''Physical Line('Periodic_4') = {'''
   for n in range(len(lower)-1):
    strc += str(lower[n]) + ','
   strc += str(lower[-1]) + '};\n'
   store.write(strc)
  else:
   for k in lower:
    additional_boundary.append(k)
   for k in upper:
    additional_boundary.append(k)

  #Collect Wall
  if argv.setdefault('inclusion',False):
   interface_surfaces = pore_wall
   boundary_surfaces = additional_boundary
  else:
   boundary_surfaces = pore_wall + additional_boundary
   interface_surfaces = []


  #Interface surface
  if len(boundary_surfaces)> 0:
   strc = r'''Physical Line('Boundary') = {'''
   for r,region in enumerate(boundary_surfaces) :
    strc += str(region)
    if r == len(boundary_surfaces)-1:
     strc += '};\n'
     store.write(strc)
    else :
     strc += ','


  if len(interface_surfaces)> 0:
   strc = r'''Physical Line('Interface') = {'''
   for r,region in enumerate(interface_surfaces) :
    strc += str(region)
    if r == len(interface_surfaces)-1:
     strc += '};\n'
     store.write(strc)
    else :
     strc += ','


 
  for n in range(len(hot)):
   periodic = FindPeriodic(hot[n],cold,points,lines)
   strc = 'Periodic Line{' + str(hot[n]) + '}={' + str(periodic) + '};\n'
   store.write(strc)

  for n in range(len(lower)):
   periodic = FindPeriodic(lower[n],upper,points,lines)
   strc = 'Periodic Line{' + str(lower[n]) + '}={' + str(periodic) + '};\n'
   store.write(strc)

#-------------------------------------------------------
  store.close()
#for p in pp:
# points.append(p)
# store.write( 'Point('+str(len(points)-1) +') = {' + str(p[0]) +','+ str(p[1])+',0,'+ str(mesh_ext) +'};\n')

#Lines--------
#for n in range(len(pp)):
# p1 = len(points) + n - len(pp)
# p2 = len(points) + (n+1)%len(pp) - len(pp)
# lines.append([p1,p2])
# store.write( 'Line('+str(len(lines)-1) +') = {' + str(p1) +','+ str(p2)+'};\n')

#Line loops
#loops += 1
#strc = 'Line Loop(' + str(loops-1)+ ') = {'
#for n in range(len(pp)):
# strc +=str(len(lines)-len(pp)+n)
# if n == len(pp)-1:
#  strc += '};\n'
# else:
#  strc += ','
#store.write(strc)

#Pores surfaces--
#ss +=1
#strc = 'Plane Surface(' + str(ss-1)+ ') = {' + str(loops-1) + '};\n'
#store.write(strc)

#plot_region(points)

#show()
#strc = r'''Physical Surface('Pores') = {'''
#for n in range(ss):
# strc += str(n)
# if n == ss-1:
#   strc += '};\n'
# else:
#   strc += ','
#store.write(strc)



from shapely.ops import cascaded_union
from shapely.geometry import Point,LineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
import shapely
import numpy as np

class Porous(object):

 def __init__(self,**argv):

  self.step = argv['step']
  self.points = []
  self.lines = []
  self.loops = []
  self.surfaces = []
  self.argv = argv
 
    
  self.common(z=-self.argv['lz']/2)
  self.common(z=self.argv['lz']/2)
  self.merge() 
  self.points = np.array(self.points)
  self.lines = np.array(self.lines)
  self.loops = np.array(self.loops)

  
  self.apply_periodic_mesh()
  
  self.write_geo()



 #def get_surface_normal(self,ss):
     
  # s = self.surfaces[ss][0]  
   #p1 = self.points[self.lines[self.loops[s,0]]]
  # p1 = self.points[self.lines[abs(self.loops[s,0])-1]]
  # p2 = self.points[self.lines[abs(self.loops[s,1])-1]]
  # normal = np.cross(p1[1]-p1[0],p2[1]-p2[0])
  # normal /= np.linalg.norm(normal)
   #if ss==2:
   #    print(normal)
   #    print(p1)
  # return normal
     
 def get_surface_centroid(self,ss):
     
  
   s = self.surfaces[ss][0]  
    
   pts = []
   for n in self.loops[s]:
    if n > 0:
        i = 0
    else:
        i = 1
    pts.append(self.points[self.lines[abs(n)-1][i]])
   pts = np.array(pts) 
    
   
   
   return np.mean(pts,axis=0)
   
     

 def isperiodic(self,s1,s2):

   #n = self.get_surface_normal(s1)
   #n2 = self.get_surface_normal(s2)
   c1 = self.get_surface_centroid(s1)
   c2 = self.get_surface_centroid(s2)
   
   
   
   if abs(np.linalg.norm(c1-c2) - self.argv['lx'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[1,0,0])) < 1e-4: 
      return 0,c1-c2
   

   if abs(np.linalg.norm(c1-c2) - self.argv['ly'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[0,1,0])) < 1e-4: 
      return 1,c1-c2
   
   if abs(np.linalg.norm(c1-c2) - self.argv['lz'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[0,0,1])) < 1e-4: 
      return 2,c1-c2
   
  
    
   return None,None
   
 
   
    
 def get_points_from_surface(self,s1):
      
   ll = self.surfaces[s1][0]
   pts = []
   ind = []
   for n in self.loops[ll]:
    if n > 0:
        i = 0
    else:
        i = 1
    pts.append(self.points[self.lines[abs(n)-1][i]])
    
    ind.append(self.lines[abs(n)-1][i])


   return np.array(pts),ind
     
     
 def apply_periodic_mesh(self):
     self.argv.setdefault('Periodic',[True,True,False])
     #Find periodic surfaces----
     self.periodic = {}
     for s1 in range(len(self.surfaces)):
      for s2 in range(s1+1,len(self.surfaces)):
         (a,per) = self.isperiodic(s1,s2)
         if not a == None and self.argv['Periodic'][a]:
          
           corr = {}
           pts1,ind1 = self.get_points_from_surface(s1)
           pts2,ind2 = self.get_points_from_surface(s2)
          
           for n1,p1 in zip(ind1,pts1):
            for n2,p2 in zip(ind2,pts2):
             if np.linalg.norm(p1-p2-per) < 1e-3:
                corr[n1] = n2
                break            
           #-----------------------------------------
           loop_1 = self.loops[self.surfaces[s1][0]] 
           loop_2 = []
           for ll in loop_1 :
             pts = self.lines[abs(ll)-1]
             if ll > 0:
              p1 = pts[0]; p2 = pts[1] 
             else:
              p1 = pts[1]; p2 = pts[0]   
             loop_2.append(self.line_exists_ordered(corr[p1],corr[p2]))             
           self.periodic[s1] = [s2,loop_1,loop_2,np.array(per/np.linalg.norm(per))]  
           
          
           
           break  
     
      
        
        

       

 def merge(self):

   #Create lines
   Nh = int(len(self.points)/2)
   for n in range(Nh):
     self.lines.append([n,n+Nh])  
     

   #Create new loops 
   N = int(len(self.loops)/2)
   for ll in range(N):
    for l in self.loops[ll]:         
     
     i1 = self.lines[l-1][0];  i2 = self.lines[l-1][1]
              
     j1 = i1 + Nh; j2 = i2 + Nh
     tmp = []
     tmp.append(self.line_exists_ordered(i1,i2))
     tmp.append(self.line_exists_ordered(i2,j2))   
     tmp.append(self.line_exists_ordered(j2,j1))
     tmp.append(self.line_exists_ordered(j1,i1))
  
         
     self.loops.append(tmp)
     self.surfaces.append([len(self.loops)-1])
     
 
 
 #def apply_periodic_mesh(self):

 def write_geo(self):

   
   store = open('mesh.geo', 'w+')

   
   #Write points-----
   for p,point in enumerate(self.points):
    store.write( 'Point('+str(p) +') = {' + str(point[0]) +','+ str(point[1])+',' + str(point[2])+','+ str(self.step) +'};\n')
   
   #Write lines----
   for l,line in enumerate(self.lines):
     store.write( 'Line('+str(l+1) +') = {' + str(line[0]) +','+ str(line[1])+'};\n')
      
   #Write loops------
   nl = len(self.lines)+1
   for ll,loop in enumerate(self.loops):  
    strc = 'Line Loop(' + str(ll+nl)+ ') = {'
    for n,l in enumerate(loop):
     strc +=str(l)
     if n == len(loop)-1:
      strc += '};\n'
     else:
      strc += ','
    store.write(strc)
   
   
   #Write surface------
   nll = len(self.loops)
   for s,surface in enumerate(self.surfaces):
    strc = '''Plane Surface(''' +str(s+nll+nl) + ''')= {''' 
    for n,ll in enumerate(surface):
      strc += str(ll+nl)  
      if n == len(surface)-1:
       strc += '};\n'
      else:
       strc += ','    
    store.write(strc)
    

   boundary = list(range(len(self.surfaces)))
   #Apply periodicity
   vec = {}
   for key, value in self.periodic.items():
      
     boundary.remove(key)  
     boundary.remove(value[0])  

     
     store.write('Periodic Surface ' + str(key+nl+nll) + ' {') 
     
     for n,k in enumerate(value[1]):
      store.write(str(k))
      if n < len(value[1])-1:
       store.write(',')
     store.write('} = ' + str(value[0]+nl+nll) + ' {')
     
     for n,k in enumerate(value[2]):
      store.write(str(k))
      if n < len(value[2])-1:
       store.write(',')
     store.write('};\n')
    

     value[0] += nl+nll
     tt = key +  nl + nll
     if np.array_equal(np.array([1,0,0]),value[3]) : 
        vec.setdefault(1,[]).append(value[0]) 
        vec.setdefault(2,[]).append(tt) 
     
     if np.array_equal(np.array([-1,0,0]),value[3]): 
        vec.setdefault(1,[]).append(tt) 
        vec.setdefault(2,[]).append(value) 
        
     if np.array_equal(np.array([0,1,0]),value[3]) : 
        vec.setdefault(3,[]).append(value[0]) 
        vec.setdefault(4,[]).append(tt)
        
     if np.array_equal(np.array([0,-1,0]),value[3]) : 
        vec.setdefault(4,[]).append(value[0]) 
        vec.setdefault(3,[]).append(tt)
        
     if np.array_equal(np.array([0,0,1]),value[3]) : 
        vec.setdefault(5,[]).append(value[0]) 
        vec.setdefault(6,[]).append(tt)   
        
     if np.array_equal(np.array([0,0,-1]),value[3]) : 
        vec.setdefault(6,[]).append(value[0]) 
        vec.setdefault(5,[]).append(tt) 
        
   for key, value in vec.items():
    
     strc = r'''Physical Surface("Periodic_''' + str(key) + '''") = {'''
     for n,p in enumerate(value) :
      strc +=str(p)
      if n == len(value)-1:
       strc += '};\n'
      else :
       strc += ','
     store.write(strc)
         
   strc = r'''Physical Surface("Boundary") = {'''
   for n,p in enumerate(boundary) :
      strc +=str(p+nl+nll)
      if n == len(boundary)-1:
       strc += '};\n'
      else :
       strc += ','
   store.write(strc)
    
    
   if len(self.surfaces) > 0:
    #------------------
    ss  = len(self.surfaces)
    strc = r'''Surface Loop(''' + str(ss+nll+nl) + ') = {'
    for s in range(len(self.surfaces)) :
     strc +=str(s+nl+nll)
     if s == len(self.surfaces)-1:
      strc += '};\n'
     else :
      strc += ','
    store.write(strc)
    
    strc = r'''Volume(0) = {''' + str(ss+nll+nl) + '};\n'
    store.write(strc)
    strc = r'''Physical Volume('Bulk') = {0};'''
    store.write(strc) 

   store.close()

 def line_exists_ordered(self,p1,p2):
     
  l = [p1,p2]   
  for n,line in enumerate(self.lines) :
   if (line[0] == l[0] and line[1] == l[1]) :
    return n+1
   if (line[0] == l[1] and line[1] == l[0]) :
    return -(n+1)
  return 0

 def already_included(self,p):

  for n,point in enumerate(self.points):
   d = np.linalg.norm(np.array(point)-np.array(p))
   if d < 1e-12:
    return n
  return -1


 def create_lines_and_loop_from_points(self,pp,z):
     
   p_list = []  
   for p in pp:
    p = [p[0],p[1],z]   
    tmp = self.already_included(p) 
    if tmp == -1:
      self.points.append(p)
      p_list.append(len(self.points)-1)
    else:
      p_list.append(tmp)


   line_list = []
   for l in range(len(p_list)):
    p1 = p_list[l]
    p2 = p_list[(l+1)%len(p_list)]
    if not p1 == p2:
     tmp = self.line_exists_ordered(p1,p2)
     if tmp == 0 : #craete line
      self.lines.append([p1,p2])
      line_list.append(len(self.lines))
     else:
      line_list.append(tmp)
      
   self.loops.append(line_list)  
      



 def common(self,z):

  Frame = Polygon(self.argv['frame'])
  polygons = self.argv['polygons']

  polypores = []
  for poly in polygons:
   thin = Polygon(poly).intersection(Frame)
   if isinstance(thin, shapely.geometry.multipolygon.MultiPolygon):
     tmp = list(thin)
     for t in tmp:
      polypores.append(t)
   else:
    polypores.append(thin)
  MP = MultiPolygon(polypores)
  bulk = Frame.difference(cascaded_union(MP))
  if not (isinstance(bulk, shapely.geometry.multipolygon.MultiPolygon)):
   bulk = [bulk]

  
  loop_start = len(self.loops)  
  for region in bulk:
   pp = list(region.exterior.coords)[:-1]
   self.create_lines_and_loop_from_points(pp,z)
   
    
   #From pores 
   for interior in region.interiors:
    pp = list(interior.coords)[:-1]
    self.create_lines_and_loop_from_points(pp,z)
  loop_end = len(self.loops)  
  
  self.surfaces.append(range(loop_start,loop_end))

    







 

  







from __future__ import absolute_import

#import os
import numpy as np
from scipy.sparse import csc_matrix
from mpi4py import MPI
#import shutil
from scipy.sparse.linalg import splu
import matplotlib.pylab as plt
from scipy.sparse import spdiags
from scipy.sparse import diags
import scipy.io
import deepdish as dd
from .geometry import Geometry
from scipy import interpolate
import sparse
from scipy.sparse import coo_matrix
import time
from numpy.testing import assert_array_equal
import pickle
import os

def log_interp1d(xx, y, kind='linear'):
     yy = y.copy()
     
     scale = min(yy)
     yy -= min(yy)
     yy+= 1e-12
     logx = np.log10(xx)
     logy = np.log10(yy)
     lin_interp = interpolate.interp1d(logx,logy,kind=kind,fill_value='extrapolate')
     log_interp = lambda zz: np.power(10.0, lin_interp(np.log10(zz)))  +scale -1e-12
     return log_interp


class Solver(object):

  def __init__(self,**argv):
   
   self.mesh = Geometry(type='load',filename = argv.setdefault('geometry_filename','geometry'))
   self.ms_error = argv.setdefault('ms_error',0.1)
   self.verbose = argv.setdefault('verbose',True)
   self.n_elems = len(self.mesh.elems)
   self.cache = os.getcwd() + '/.cache'
   self.multiscale = argv.setdefault('multiscale',False)
   tmp = dd.io.load('material.hdf5')
   
   if self.mesh.dim == 3:
     self.mat = tmp['data_3D']
   else:   
     self.mat = tmp['data_2D']
       
   self.kappa_factor = self.mesh.kappa_factor
   if self.verbose: 
    self.print_logo()
    self.print_dof()


   #Compute directional connections-------
   if self.mesh.dim == 3:   
    self.n_index = self.mat['n_phi'] * self.mat['n_theta']
   else:   
    self.n_index = self.mat['n_phi']

   start = time.time() 
   self.compute_directional_connections()  
   #print(time.time()-start)

   #quit() 
    
   
   #quit()
   #------------------------- 
   if self.verbose: self.print_bulk_kappa()
   
   self.assemble_fourier()
   
   #---------------------------
   
   #solve the BTE
   self.solve_bte(**argv)
   


   #if MPI.COMM_WORLD.Get_rank() == 0:
   # if os.path.isdir(self.cache):
   #  shutil.rmtree(self.cache)

   #SAVE FILE--------------------
   #if argv.setdefault('save',True):
   # if MPI.COMM_WORLD.Get_rank() == 0:
   #  dd.io.save('solver.hdf5', self.state)




  def compute_directional_connections(self):

  
   start = time.time() 
   comm = MPI.COMM_WORLD
   block = self.n_index // comm.size + 1
   for kk in range(block):
    index = comm.rank*block + kk   
    if index < self.n_index:   
 
     if self.mesh.dim == 3:   
      t = int(index/self.mat['n_phi'])
      p = index%self.mat['n_phi']   
      angle = self.mat['direction_ave'][t][p]
     else: 
      angle = self.mat['polar_ave'][index]   
    
     r=[];c=[];d=[]
     rk=[]; ck=[]; dk=[]
     dk2=[]
     P = np.zeros(self.n_elems)
     HW_minus = np.zeros(self.n_elems)
     HW_plus = np.zeros(self.n_elems)
     for i,j in zip(*self.mesh.A.nonzero()):

      (cm,cp,cmb,cpb) = self.mesh.get_angular_coeff_old(i,j,angle)

      r.append(i); c.append(i); d.append(cp) 
      r.append(i); c.append(j); d.append(cm)
      v = self.mesh.get_side_periodic_value(i,j)
      rk.append(i); ck.append(j);
      dk.append(v*(cp+cm)*self.mesh.get_elem_volume(i)*0.5)
      P[i] += v*cm
      HW_plus[i]  += cpb
      HW_minus[i] -= cmb

    A = csc_matrix( (d,(r,c)), shape=(self.n_elems,self.n_elems))    
    K = csc_matrix( (dk,(rk,ck)), shape=(self.n_elems,self.n_elems) )

    #--------------------------------------------------------------
    #--------------------------WRITE FILES---------------------
    A = csc_matrix( (d,(r,c)), shape=(self.n_elems,self.n_elems) )
    K = csc_matrix( (dk,(rk,ck)), shape=(self.n_elems,self.n_elems) )
    scipy.io.mmwrite(self.cache + '/A_' + str(index) + r'.mtx',A)
    scipy.io.mmwrite(self.cache + '/K_' + str(index) + r'.mtx',K)
    P.dump(open(self.cache + '/P_' + str(index) +r'.np','wb+'))
    HW_minus.dump(open(self.cache +'/HW_MINUS_' + str(index) +r'.np','wb+'))
    HW_plus.dump(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','wb+'))


  

      
  def solve_bte(self,**argv):

   if not argv.setdefault('load_state',False):
    argv.update({'kappa':[self.mat['kappa_bulk_tot']]})
    output = self.solve_fourier(**argv) #simple fourier----   
    temperature_fourier = np.tile(output['temperature_fourier'],(self.mat['n_mfp'],1))
    #temperature_fourier_gradient = np.tile(output['temperature_fourier_gradient'],(self.n_rmfp,1,1))
    #ms = np.zeros((self.n_mfp,self.n_theta,self.n_phi))
    TB = temperature_fourier.copy()
    TL = temperature_fourier.copy()     
    error = 1.0
    kappa_old = 0
    n_iter = 0
    #SUP = output['kappa_fourier'][0]*np.ones(len(self.mat['J3']))
   #kappa_fourier =  np.tile(output['kappa_fourier'],(self.mat['n_mfp']))
   else:
    data = dd.io.load('state.hdf5')
    TL = data['TL']
    TB = data['TB']
    SUP = data['SUP']
    n_iter = data['n_iter']
    kappa_old = data['kappa']
    kappa = data['kappa']
    
   
    
   #FLUX = np.zeros((self.n_el,3))
   
   comm = MPI.COMM_WORLD  
   block = self.n_index // comm.size + 1
   lu = {}

   rank = comm.rank 

  
   while n_iter < argv.setdefault('max_bte_iter',10): #and \
#          error > argv.setdefault('max_bte_error',1e-2):

      
              
    #solve_fourier
    #if n_iter ==  argv.setdefault('max_bte_iter',10)-1:
     #Solve Fourier-----------------------------------------------
    #if n_iter == argv.setdefault('max_bte_iter',10)-1:
     #argv.update({'mfe_factor':1.0,\
     #             'verbose':False,\
     #             'lattice_temperature':TL.copy(),\
     #             'boundary_temperature':TB.copy(),\
     #             'kappa':self.mat['kappa_mfe'],\
     #             'interface_conductance': self.mat['G']})    
     #output = self.solve_fourier(**argv) 
     #kappa_fourier = output['kappa_fourier'] 
    
    
    Jp,J = np.zeros((2,self.mat['n_mfp'],self.n_elems))
    kernelp,kernel = np.zeros((2,self.mat['n_mfp']))
    Tp,T = np.zeros((2,self.mat['n_mfp'],self.n_elems))
    for kk in range(block):
     index = rank*block + kk   
     if index < self.n_index  :
      #Read directional information-------------------------------------------------- 
      A = scipy.io.mmread(self.cache + '/A_' + str(index) + '.mtx').tocsc()
      P = np.load(open(self.cache + '/P_' + str(index) +r'.np','rb'))
      HW_MINUS = np.load(open(self.cache + '/HW_MINUS_' + str(index) +r'.np','rb'))
      HW_PLUS = np.load(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','rb'))
      K = scipy.io.mmread(self.cache + '/K_' + str(index) + '.mtx').tocsc()
       
      for n in range(self.mat['n_mfp'])[::-1]:
          
       local_index = index*self.mat['n_mfp'] + n   
       #Solve the system--------    
       RHS = self.mat['mfp'][n]  * (P + np.multiply(TB[n],HW_MINUS)) + TL[n]      
       #if not local_index in lu.keys():   
       F = scipy.sparse.eye(self.n_elems) +  A * self.mat['mfp'][n] 
       lu = splu(F.tocsc())
       temp = lu.solve(RHS)
       #lu.update({local_index:splu(F.tocsc())})   
       #temp = lu[local_index].solve(RHS)        
       
       Tp[n] += temp*self.mat['domega'][index]
       kernelp[n] += 3*self.kappa_factor*K.dot(temp-TL[n]).sum()*self.mat['domega'][index]
       Jp[n] += np.multiply(temp,HW_PLUS)*self.mat['domega'][index]
    
    
    comm.Allreduce([Tp,MPI.DOUBLE],[T,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([kernelp,MPI.DOUBLE],[kernel,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([Jp,MPI.DOUBLE],[J,MPI.DOUBLE],op=MPI.SUM)
    
    n_iter +=1
   
    #----------------------------------------------------------- 

    #Suppression function   
    SUP = np.sum(np.multiply(self.mat['J3'],log_interp1d(self.mat['mfp'],kernel.clip(min=1e-13))(self.mat['trials'])),axis=1)
    kappa = np.dot(self.mat['kappa_bulk'],SUP)
    error = abs(kappa-kappa_old)/abs(kappa)
    kappa_old = kappa
    
    if argv.setdefault('save_state',False) and rank == 0:
     dd.io.save('state.hdf5',{'TL':TL,'TB':TB,'SUP':SUP,'n_iter':n_iter,'error':error,'kappa':kappa})

    #Boundary temperature
    TB = 4.0*np.tile([np.sum(np.multiply(self.mat['J1'],log_interp1d(self.mat['mfp'],J.T[e])(self.mat['trials']))) \
         for e in range(self.n_elems)],(self.mat['n_mfp'],1))
    
    #Lattice temperature   
    TL = np.tile([np.sum(np.multiply(self.mat['J2'],log_interp1d(self.mat['mfp'],T.T[e])(self.mat['trials']))) \
        for e in range(self.n_elems)],(self.mat['n_mfp'],1))
    
    
    
       
    #Thermal conductivity   
    if rank==0:
     print(kappa)       
     

    #if n_iter ==  argv.setdefault('max_bte_iter',10)-1:
     #Solve Fourier-----------------------------------------------
     #if n_iter == argv.setdefault('max_bte_iter',10)-1:
   #argv.update({'mfe_factor':1.0,\
   #               'verbose':False,\
   #               'lattice_temperature':TL.copy(),\
   #               'boundary_temperature':TB.copy(),\
   #               'kappa':self.mat['kappa_mfe'],\
   #               'interface_conductance': self.mat['G']})    
   #output = self.solve_fourier(**argv) 
   #kappa_fourier = output['kappa_fourier'] 
    

   if rank==0:
     print(' ')  
     print(kappa)       

    #if self.multiscale:
    # SUP_DIF = np.sum(np.multiply(self.mat['J0'],log_interp1d(self.mat['mfp'],kappa_fourier/self.mat['kappa_mfe'])(self.mat['trials'])),axis=1)   

    # plt.plot(self.mat['mfp_bulk'],SUP_DIF,'r') 
    # plt.plot(self.mat['mfp_bulk'],SUP) 
    #
    # plt.xscale('log')
    # plt.show()
    
   
    
    
    
    
    

  def print_bulk_kappa(self):
      
    if MPI.COMM_WORLD.Get_rank() == 0:
     if self.verbose:
      #print('  ')
      #for region in self.region_kappa_map.keys():
       print('{:8.2f}'.format(self.mat['kappa_bulk_tot'])+ ' W/m/K')    
       #print(region.ljust(15) +  '{:8.2f}'.format(self.region_kappa_map[region])+ ' W/m/K')    

   
   
   

  def assemble_fourier(self) :

   if  MPI.COMM_WORLD.Get_rank() == 0:
    row_tmp = []
    col_tmp = []
    data_tmp = []
    row_tmp_b = []
    col_tmp_b = []
    data_tmp_b = []
    data_tmp_b2 = []
    B = np.zeros(self.n_elems)
    kappa_mask = []
    for kc1,kc2 in zip(*self.mesh.A.nonzero()):
     #ll = self.mesh.get_side_between_two_elements(kc1,kc2)

     #kappa = self.side_kappa_map[ll]
     kappa = 1.0
     #kappa = 100.0
     #print(kappa)        print(KAPPAp)


     (v_orth,dummy) = self.mesh.get_decomposed_directions(kc1,kc2)
     vol1 = self.mesh.get_elem_volume(kc1)
     #vol2 = self.mesh.get_elem_volume(kc2)
     row_tmp.append(kc1)
     col_tmp.append(kc2)
     data_tmp.append(-v_orth/vol1*kappa)
     row_tmp.append(kc1)
     col_tmp.append(kc1)
     data_tmp.append(v_orth/vol1*kappa)
     #----------------------
     row_tmp_b.append(kc1)
     col_tmp_b.append(kc2)
     tt = self.mesh.get_side_periodic_value(kc2,kc1)
     data_tmp_b.append(tt*v_orth*kappa)
     data_tmp_b2.append(tt)
     if tt == 1.0:
      kappa_mask.append([kc1,kc2])
     
     #---------------------
     B[kc1] += self.mesh.get_side_periodic_value(kc2,kc1)*v_orth/vol1*kappa

    #Boundary_elements
    FF = np.zeros(self.n_elems)
    boundary_mask = np.zeros(self.n_elems)
    for side in self.mesh.side_list['Boundary']:
     elem = self.mesh.side_elem_map[side][0]
     vol = self.mesh.get_elem_volume(elem)
     area = self.mesh.get_side_area(side)
     FF[elem] = area/vol
     boundary_mask[elem] = 1.0

    F = csc_matrix((np.array(data_tmp),(np.array(row_tmp),np.array(col_tmp))), shape=(self.n_elems,self.n_elems) )
    RHS = csc_matrix( (np.array(data_tmp_b),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )
    PER = csc_matrix( (np.array(data_tmp_b2),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )

    #-------
    data = {'F':F,'RHS':RHS,'B':B,'PER':PER,'boundary_mask':boundary_mask,'FF':FF,'kappa_mask':kappa_mask}
   else: data = None
   data =  MPI.COMM_WORLD.bcast(data,root=0)
   self.F = data['F']
   self.RHS = data['RHS']
   self.PER = data['PER']
   self.kappa_mask = data['kappa_mask']
   self.B = data['B']
   self.FF = data['FF']
   self.boundary_mask = data['boundary_mask']


  def solve_fourier(self,**argv):
       
    rank = MPI.COMM_WORLD.rank    
    #Update matrices---------------------------------------------------
    G = argv.setdefault('interface_conductance',np.zeros(1))#*self.dom['correction']
    TL = argv.setdefault('lattice_temperature',np.zeros((1,self.n_elems)))
    TB = argv.setdefault('boundary_temperature',np.zeros((1,self.n_elems)))
    mfe_factor = argv.setdefault('mfe_factor',0.0)
    kappa_bulk = argv.setdefault('kappa',[1.0])
   
    
    
    #print(kappa_bulk)
    #G *=5.0
    n_index = len(kappa_bulk)

    #G = np.zeros(n_index)
    #-------------------------------
    KAPPAp,KAPPA = np.zeros((2,n_index))
    FLUX,FLUXp = np.zeros((2,n_index,3,self.n_elems))
    TLp,TLtot = np.zeros((2,n_index,self.n_elems))

    comm = MPI.COMM_WORLD  
    size = comm.size
    rank = comm.rank
    block = n_index // size + 1
    for kk in range(block):
     index = rank*block + kk   
     if index < n_index:
    
    #set up MPI------------------
    #for index in range(n_index):
      #Aseemble matrix-------------       
     
      A  = kappa_bulk[index] * self.F + mfe_factor * csc_matrix(scipy.sparse.eye(self.n_elems)) 
      A += mfe_factor * spdiags(self.FF,0,self.n_elems,self.n_elems) * G[index]  
      
      B  = kappa_bulk[index] * self.B + mfe_factor * TL[index] 
      B += mfe_factor * np.multiply(self.FF,TB[index]) * G[index]
      
      if mfe_factor == 0.0: A[10,10] = 1.0
      SU = splu(A)
      C = np.zeros(self.n_elems)
      #--------------------------------------

      n_iter = 0
      kappa_old = 0
      error = 1        

      
      while error > argv.setdefault('max_fourier_error',1e-2) and \
                    n_iter < argv.setdefault('max_fourier_iter',10) :
                    
        RHS = B + C*kappa_bulk[index]
        if mfe_factor == 0.0: RHS[10] = 0.0      
        temp = SU.solve(RHS)
        temp = temp - (max(temp)+min(temp))/2.0
        (C,flux) = self.compute_non_orth_contribution(temp)
        
        KAPPAp[index] = self.compute_diffusive_thermal_conductivity(temp)*kappa_bulk[index]
        error = abs((KAPPAp[index] - kappa_old)/KAPPAp[index])
        
        kappa_old = KAPPAp[index]
        n_iter +=1
        
      #FLUXp = np.array([-self.elem_kappa_map[k]*tmp for k,tmp in enumerate(flux)])
      FLUXp[index] = flux.T
      TLp[index] = temp
     
        
    comm.Allreduce([KAPPAp,MPI.DOUBLE],[KAPPA,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([TLp,MPI.DOUBLE],[TLtot,MPI.DOUBLE],op=MPI.SUM) #to be changed
    comm.Allreduce([FLUXp,MPI.DOUBLE],[FLUX,MPI.DOUBLE],op=MPI.SUM)  
      
    
    
    if argv.setdefault('verbose',True) and rank==0:

     print('  ')
     print('Thermal Conductivity: '.ljust(20) +  '{:8.2f}'.format(KAPPAp.sum())+ ' W/m/K')
     print('  ')
    


    return {'kappa_fourier':KAPPA,'temperature_fourier_gradient':FLUX,'temperature_fourier':TLtot}

      
  def compute_non_orth_contribution(self,temp) :

    self.gradT = self.mesh.compute_grad(temp)
    C = np.zeros(self.n_elems)
    for i,j in zip(*self.mesh.A.nonzero()):

     if not i==j:
      #Get agerage gradient----
      side = self.mesh.get_side_between_two_elements(i,j)


      w = self.mesh.get_interpolation_weigths(side,i)
      grad_ave = w*self.gradT[i] + (1.0-w)*self.gradT[j]
      #------------------------
      (dumm,v_non_orth) = self.mesh.get_decomposed_directions(i,j)

      
      C[i] += np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(i)#*kappa
      C[j] -= np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(j)#*kappa


    return C, self.gradT# [-self.elem_kappa_map[n]*tmp for n,tmp in enumerate(self.gradT)]


  def compute_inho_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   kappa = 0
   for i,j in zip(*self.PER.nonzero()):

    #ll = self.mesh.get_side_between_two_elements(i,j)
    #kappa_b = self.side_kappa_map[ll]
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
   
    
    kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])#*kappa_b

   return kappa*self.kappa_factor


  def compute_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   
   kappa = 0
   for (i,j) in self.kappa_mask:
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
    kappa += v_orth  *(temp[j]+1.0-temp[i])

   #kappa = 0
   #for i,j in zip(*self.PER.nonzero()):
   # (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
   # kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])

   return kappa*self.kappa_factor

  def print_logo(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    print(' ')
    print(r'''  ___                   ____ _____ _____ ''' )
    print(r''' / _ \ _ __   ___ _ __ | __ )_   _| ____|''')
    print(r'''| | | | '_ \ / _ \ '_ \|  _ \ | | |  _|  ''')
    print(r'''| |_| | |_) |  __/ | | | |_) || | | |___ ''')
    print(r''' \___/| .__/ \___|_| |_|____/ |_| |_____|''')
    print(r'''      |_|                                ''')
    print('')
    print('Giuseppe Romano [romanog@mit.edu]')
    print(' ')

  def print_dof(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    if self.verbose:
     print(' ')
     print('Elements:          ' + str(self.n_elems))
     print('Azimuthal angles:  ' + str(self.mat['n_theta']))
     print('Polar angles:      ' + str(self.mat['n_phi']))
     print('Mean-free-paths:   ' + str(self.mat['n_mfp']))
     



import numpy as np
import os


   
   
def compute_dom_2d(argv):
    
   #Polar Angle-----
   n_phi = int(argv.setdefault('n_phi',48)); Dphi = 2.0*np.pi/n_phi
   phi = np.linspace(0.0,2.0*np.pi,n_phi,endpoint=False)
   phi += argv.setdefault('polar_offset',0.0)*np.pi/180.0
   phi = phi %(2.0*np.pi)
   fphi= np.sinc(Dphi/2.0/np.pi)
   dphi_ave = 1/n_phi*np.ones(n_phi)
   #--------------------
   
   #Azimuthal Angle------------------------------
   n_theta = int(argv.setdefault('n_theta',24)); 
   Dtheta = np.pi/n_theta/2.0
   theta = np.linspace(Dtheta/2.0,np.pi/2.0 - Dtheta/2.0,n_theta)
   Dtheta_int = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)
   
   
   ftheta = np.sin(theta)*(1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
   dtheta_ave = np.sin(Dtheta/2.0)*np.sin(theta) 
   dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)
   #---------------------------------------------
   
   #phonon directions-------
   polar_dir = np.array([np.sin(phi),np.cos(phi),np.zeros(n_phi)]).T
   
   
   output = {'n_phi':n_phi,\
             'dtheta':dtheta,\
             'n_theta':n_theta,\
             'dphi_ave':dphi_ave,\
             'dphi':Dphi,\
             'dtheta_ave':dtheta_ave,\
             'polar_dir':polar_dir,\
             'fphi':fphi,\
             'ftheta':ftheta}
   
   return output



#def compute_dom_3d(argv) :

   #print(argv.setdefault('n_theta',16))


#   n_theta = int(argv.setdefault('n_theta',16))
#   n_phi = int(argv.setdefault('n_phi',48))
   #n_phi = int(argv['n_phi'])


   #apply_symmetry = False
#   norm_dom = 4.0*np.pi
#   output = {}
#   Dtheta = np.pi/n_theta
#   d_theta_plain = Dtheta*np.ones(n_theta)
#   theta_vec = np.linspace(Dtheta/2.0,np.pi - Dtheta/2.0,n_theta)
 #  output.update({'theta_vec':theta_vec})
 #  output.update({'n_theta':n_theta})
   #---------------------------------------

   #Phi ----------------------------------
 #  Dphi = 2.0*np.pi/n_phi
 #  d_phi_plain = Dphi*np.ones(n_phi)
 #  d_phi_int = Dphi*np.ones(n_phi)
 #  output.update({'dphi':Dphi})
   
 #  output.update({'fphi':2.0*np.sin(Dphi/2.0)/Dphi})   #
   

   #phi_vec = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True) #this is the correct one

 #  phi_vec = np.linspace(0.0,2.0*np.pi,n_phi,endpoint=False)
 #  phi_vec += argv.setdefault('polar_offset',0.0)*np.pi/180.0
 #  phi_vec = phi_vec %(2.0*np.pi)


 #  output.update({'phi_vec':phi_vec})
 #  output.update({'d_phi_vec':d_phi_plain})
 #  output.update({'n_phi':n_phi})
 #  output.update({'n_theta':n_theta})
 #  output.update({'d_phi':Dphi})
   #-----------------------------------------------------------------

 #  d_theta_int = np.zeros(n_theta)
 #  ss = np.zeros((n_theta,n_phi,3,3))
#   d_theta_vec = np.zeros(n_theta)
 #  phonon_dir = np.zeros((n_theta,n_phi,3))
 #  integrated_dir = np.zeros((n_theta,n_phi,3))
 #  phi_dir = np.zeros((n_phi,3))
 #  polar_dir = np.zeros((n_phi,3))

 #  at = np.zeros(n_theta)
 #  tt = np.zeros(n_theta)
 #  for t in range(n_theta):
 #   theta = theta_vec[t]

  #  int_sin_theta = 2.0*np.sin(d_theta_plain[t]/2.0)*np.sin(theta)
  #  int_sin_theta_2 = 1.0/2.0*(d_theta_plain[t]-np.cos(2.0*theta)*np.sin(d_theta_plain[t]))
  #  int_sin_theta_3 = 1.0/6.0*(9.0*np.sin(theta)*np.sin(d_theta_plain[t]/2.0)-np.sin(3.0*theta)*np.sin(3.0/2.0*d_theta_plain[t])   )
  #  int_cos_theta_2_sin_theta = 1.0/6.0*(3.0*np.sin(theta)*np.sin(d_theta_plain[t]/2.0)+\
   #                                       np.sin(3.0*theta)*np.sin(3.0/2.0*d_theta_plain[t]))
  #  int_sin_theta_2_cos_theta = 1.0/3.0*(pow(np.sin(1.0/2.0*(d_theta_plain[t]-2.0*theta)),3.0) +\
   #                                      (pow(np.sin(theta+d_theta_plain[t]/2.0),3)))
   # int_cos_theta_sin_theta = np.sin(d_theta_plain[t])*np.sin(theta)*np.cos(theta)
#
#    d_theta_int[t] = int_sin_theta
 #   at[t] = int_sin_theta_2
    
    
  #  tt[t]=np.sin(theta)*(1-np.cos(2*theta)*np.sinc(d_theta_plain[t]/np.pi))/(np.sinc(d_theta_plain[t]/2/np.pi)*(1-np.cos(2*theta)))



   # for p in range(n_phi):

    # phi = phi_vec[p]
     #------------------------------------------------------
    # int_cos_phi = 2.0*np.cos(phi)*np.sin(d_phi_plain[p]/2.0)
    # int_sin_phi = 2.0*np.sin(phi)*np.sin(d_phi_plain[p]/2.0)
    # int_sin_phi_2 = 1.0/2.0*(d_phi_plain[p]-np.cos(2.0*phi)*np.sin(d_phi_plain[p]))
    # int_cos_phi_2 = 1.0/2.0*(np.cos(2.0*phi)*np.sin(d_phi_plain[p])+d_phi_plain[p])
    # int_cos_phi_sin_phi = np.sin(phi)*np.cos(phi)*np.sin(d_phi_plain[p])
     #-------------------------------------------

     
    # x = sin(theta) * sin(phi)
    # y = sin(theta) * cos(phi)
    # z = cos(theta)
    # phonon_dir[t][p] = np.array([x,y,z])
    # polar_dir[p] = [round(sin(phi),8),round(cos(phi),8),0.0]

    # phi_dir[p] = [round(int_sin_phi,8),round(int_cos_phi,8),Dphi]
    # integrated_dir[t][p][0] = phi_dir[p][0]*at[t]
    # integrated_dir[t][p][1] = phi_dir[p][1]*at[t]
    # integrated_dir[t][p][2] = phi_dir[p][2]*int_cos_theta_sin_theta


    # ss[t][p][0][0] = int_sin_theta_3*int_sin_phi_2
    # ss[t][p][0][1] = int_sin_theta_3*int_cos_phi_sin_phi
    # ss[t][p][1][0] = ss[t][p][0][1]
    # ss[t][p][0][2] = int_sin_theta_2_cos_theta * int_sin_phi
    # ss[t][p][2][0] = ss[t][p][0][2]
    # ss[t][p][1][2] = int_sin_theta_2_cos_theta * int_cos_phi
    # ss[t][p][2][1] = ss[t][p][1][2]
    # ss[t][p][1][1] = int_sin_theta_3*int_cos_phi_2
    # ss[t][p][2][2] = int_cos_theta_2_sin_theta*d_phi_plain[p]



   #d_omega = np.outer(d_theta_int,d_phi_int)

   #ss2 = np.zeros((n_theta,n_phi,3,3))

   #tmp = np.zeros((3,3))
   #tmp2 = 0.0
   #for t in range(n_theta):
   # for p in range(n_phi):
   #  ss2[t,p] =  np.outer(integrated_dir[t,p],integrated_dir[t,p])/d_omega[t,p]
   #  tmp += ss2[t][p]
   #  if np.dot(phonon_dir[t][p],[1,0,0]) >= 0:
   #   tmp2 +=integrated_dir[t,p][0]


   #correction = 4.0*tmp2/np.trace(tmp)

   #print(correction)
   #quit()
   #for t in range(n_theta):
#    for p in range(n_phi):#
      #print(np.cross(integrated_dir[t][p][0:2],phonon_dir[t][p][0:2]))


   #output.update({'d_omega':d_omega})
   #output.update({'correction':correction})
   #output.update({'d_theta_vec':d_theta_int})
   #output.update({'ss':ss})
   #output.update({'ss2':ss2})
   #output.update({'n_theta':n_theta})
   #output.update({'n_phi':n_phi})
   #output.update({'phonon_dir':phonon_dir})
   #output.update({'integrated_dir':integrated_dir})
   #output.update({'S':integrated_dir})
   #output.update({'s':phonon_dir})
   #output.update({'norm_dom':norm_dom})
   #output.update({'phi_dir':phi_dir})
   #output.update({'polar_dir':polar_dir})
   #output.update({'at':at})


   #return output



from setuptools import setup,find_packages
import os

setup(name='openbte',
      version='0.9.38',
      description='Boltzmann Transport Equation for Phonons',
      author='Giuseppe Romano',
      author_email='romanog@mit.edu',
      classifiers=['Programming Language :: Python :: 2.7',\
                   'Programming Language :: Python :: 3.6'],
      long_description=open('README.rst').read(),
      install_requires=['numpy',
                        'scipy',
                        'sparse',
                        'shapely',
                        'pyvtk',
                        'unittest2',
                        'ipython',
                        'termcolor',
                        'future',
                        'alabaster',
                        'deepdish',
                        'mpi4py',
                        'matplotlib',
                         ],
      license='GPLv2',\
      packages = ['openbte'],
      package_data = {'openbte':['materials/*.dat','../extern/gmsh']},
      #packages = find_packages(exclude=['openbte_new']),
      entry_points = {
     'console_scripts': [
      'openbte=openbte.__main__:main','shengbte2openbte=openbte.shengbte2openbte:main'],
      },
      zip_safe=False)



from mpi4py import MPI
import numpy as np



def compute_serial_sum(func,n_tot,output,options):

  for key, value in output.items():
   strc = r'''output[' ''' + key + r''' '].fill(0)'''
   exec("".join(strc).replace(" ", ""))
  #Do the sum
  for n in range(n_tot):
   tmp = func(n,options)
   for key, value in output.items():
    strc = r'''output[' ''' + key + r''' ']+=tmp[' ''' + key +r''' '] '''
    exec(strc.replace(" ", ""))


def compute_parallel_sum(func,n_tot,output,options):

   comm = MPI.COMM_WORLD

   comm.Barrier()

   
   size = comm.Get_size()
   n_left = n_tot
   n_start = 0
   #data = {}
   #VARIABLES------------------------------
   for key, value in output.items():
    strc = r'''output[' ''' + key + r''' '].fill(0)'''
    exec("".join(strc).replace(" ", ""))
    exec(key + '_tot = value')
   #-------------------------------------
   while n_left > 0:
    n_cycle = min([n_left,size-1])
    if comm.Get_rank() > 0:
     n = n_start + comm.Get_rank()-1
     if n < n_tot:
      tmp = func(n,options)
      for key, value in output.items():
       strc = key + r''' = tmp[' ''' + key +r''' '] '''
       exec(strc.replace(" ", ""))
     else:
      for key, value in output.items():
       exec(key + ' = value')
      #----------------------------------

    else :
     for key, value in output.items():
      exec(key + ' = ' + key + '_tot.copy()')

    for key, value in output.items():
     exec('comm.Reduce([' + key+',MPI.DOUBLE],[' + key+'_tot,MPI.DOUBLE],op=MPI.SUM,root=0)')

    n_left  -= n_cycle
    n_start += n_cycle

   comm.Barrier()

   strc = r'''data = {   '''
   for key, value in output.items():
    strc += r''' ' ''' + key + r''' ':  ''' + key +'_tot,'
   strc = list(strc)

   strc[-1] = '}'
   strc = "".join(strc).replace(" ", "")
   
   exec(strc)
   #print(data)
   #print(locals().keys())
   #quit()
   tmp = comm.bcast(data,root=0)
   for key, value in tmp.items():
    if comm.Get_rank() > 0:
     strc = r'''output[' ''' + key+r''' '] +=tmp[' ''' + key+r''' ']'''
     exec("".join(strc).replace(" ", ""))



def compute_sum(func,n_tot,output = {},options = {}):
  if MPI.COMM_WORLD.Get_size() > 1 :
   compute_parallel_sum(func,n_tot,output,options)
  else:
   compute_serial_sum(func,n_tot,output,options)
